"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Collection subclass:#SequenceableCollection
	instanceVariableNames:''
	classVariableNames:''
	poolDictionaries:''
	category:'Collections-Abstract'
!

!SequenceableCollection class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

documentation
"
    SequenceableCollections have ordered elements which can be accessed via
    a numeric index. 
    SequenceableCollection is an abstract class - there are no instances of 
    it in the system. 

    The methods found here assume that implementations of (at least)
    #at:/#at:put: are implemented (which are always found in Object for
    indexable objects).

    For performance, some subclasses redefine more methods, knowing the details
    of how elements are stored. 
    Especially, bulk data manipulation (i.e. #replaceFrom:to:with:startingAt:) 
    and search methods (i.e. #indexOf:startingAt:) are good candidates for this
    kind of tuneup.

    See concrete subclasses (such as OrderedCollection, Array or String).

    [author:]
        Claus Gittinger

    [see also:]
        OrderedCollection Array
        CharacterArray String
"
! !

!SequenceableCollection class methodsFor:'instance creation'!

new:size withAll:element
    "return a new collection of size, where all elements are
     initialized to element."

    |newCollection|

    newCollection := self new:size.
    newCollection atAllPut:element.
    ^ newCollection
!

withSize:size
    "return a new collection of size.
     For variable size collections, this is different from #new:,
     in that #new: creates an empty collection with preallocated size,
     while #withSize: creates a non empty one."

    |newCollection|

    newCollection := self new:size.
    newCollection grow:size.
    ^ newCollection

    "
     (OrderedCollection new:10) inspect.
     (OrderedCollection withSize:10) inspect.
     (Array new:10) inspect.
     (Array withSize:10) inspect.
    "
! !

!SequenceableCollection methodsFor:'accessing'!

after:anObject
    "return the element, after anObject.
     If anObject is not in the receiver, report an error.
     This depends on #after:ifAbsent: being implemented in a concrete class."

    ^ self after:anObject ifAbsent:[self errorValueNotFound:anObject]

    "
     #(4 3 2 1) asOrderedCollection after:3. 
     #(4 3 2 1) asOrderedCollection after:5 
     #(4 3 2 1) asOrderedCollection after:1 
    "

    "Created: 10.5.1996 / 13:59:29 / cg"
    "Modified: 10.5.1996 / 14:03:33 / cg"
!

after:anObject ifAbsent:exceptionBlock
    "return the element, after anObject. If there is no such element,
     return the value from exceptionBlock."

    ^ self subclassResponsibility

    "Created: 10.5.1996 / 14:00:10 / cg"
!

at:index ifAbsent:exceptionBlock
    "return the element at index if valid. 
     If the index is invalid, return the result of evaluating 
     the exceptionblock.
     NOTICE: 
	in ST-80, this message is only defined for Dictionaries,
	however, having a common protocol with indexed collections
	often simplifies things."

    ((index < 1) or:[index > self size]) ifTrue:[
	^ exceptionBlock value
    ].
    ^ self at:index

    "
     #(1 2 3) at:4 ifAbsent:['no such index']
     #(1 2 3) asOrderedCollection at:4 ifAbsent:['no such index'] 

     (Dictionary with:(#foo -> #bar)
		 with:(#frob -> #baz)) 
	 at:#foobar ifAbsent:['no such index']
    "
!

before:anObject
    "return the element before the argument, anObject.
     If anObject is not in the receiver, report an error.
     This depends on #before:ifAbsent: being implemented in a concrete class."

    ^ self before:anObject ifAbsent:[self errorValueNotFound:anObject]

    "
     #(4 3 2 1) asOrderedCollection before:3. 
     #(4 3 2 1) asOrderedCollection before:4 
     #(4 3 2 1) asOrderedCollection before:0 
    "

    "Created: 10.5.1996 / 14:03:27 / cg"
    "Modified: 10.5.1996 / 14:03:44 / cg"
!

before:anObject ifAbsent:exceptionBlock
    "return the element, before anObject. If there is no such element,
     return the value from exceptionBlock."

    ^ self subclassResponsibility

    "Created: 10.5.1996 / 14:03:53 / cg"
!

first
    "return the first element"

    ^ self at:1
!

keyAtEqualValue:value
    "return the index of a value.
     This is normally not used (use indexOf:), but makes the
     protocol more compatible with dictionaries."

    ^ self indexOf:value
!

keyAtEqualValue:value ifAbsent:exceptionBlock
    "return the index of a value.
     This is normally not used (use indexOf:), but makes the
     protocol more compatible with dictionaries."

    ^ self indexOf:value ifAbsent:exceptionBlock
!

keyAtValue:value
    "return the index of a value.
     This is normally not used (use indexOf:), but makes the
     protocol more compatible with dictionaries."

    ^ self identityIndexOf:value
!

keyAtValue:value ifAbsent:exceptionBlock
    "return the index of a value.
     This is normally not used (use indexOf:), but makes the
     protocol more compatible with dictionaries."

    ^ self identityIndexOf:value ifAbsent:exceptionBlock
!

last
    "return the last element"

    ^ self at:(self size)
!

swap:index1 with:index2
    "exchange two elements"

    |t|

    t := self at:index1.
    self at:index1 put:(self at:index2).
    self at:index2 put:t

    "Modified: 15.10.1997 / 19:27:08 / cg"
!

upTo:anElement
    "return a new collection consisting of the receivers elements upTo
     (but excluding) anElement.
     If anElement is not in the receiver, the returned collection
     will consist of all elements of the receiver"

    |pos|

    pos := self indexOf:anElement.
    pos == 0 ifTrue:[^ self copy].

    ^ self copyFrom:1 to:(pos - 1)

    "
     #(1 2 3 4 5 6 7 8 9) upTo:5  
     'hello world' upTo:Character space  
     #(9 8 7 6 5 4 3 2 1) asSortedCollection upTo:5 

     raises an error:

     (Dictionary new 
        at:#foo put:'foo';
        at:#bar put:'bar';
        at:#baz put:'baz';
        yourself) upTo:#bar
                
    "

    "Modified: 11.5.1996 / 11:14:05 / cg"
! !

!SequenceableCollection methodsFor:'adding & removing'!

add:anObject
    "append the argument, anObject to the collection.
     Return the argument, anObject.

     Notice, that this is modifies the receiver NOT a copy.
     Also note, that it may be a slow operation for some collections,
     due to the grow:-message, which is inefficient for fixed size 
     collections (i.e. for Strings and Arrays it is not recommened)."

    |newSize|

    newSize := self size + 1.
    self grow:newSize.
    self at:newSize put:anObject.
    ^ anObject

    "
     |a|
     a := #(1 2 3 4 5 6 7 8).
     a add:'hello'.
     a
    "
    "
     |c|
     c := #(1 2 3 4 5 6 7 8) asOrderedCollection.
     c add:'hello'.
     c
    "
!

add:anElement beforeIndex:index
    "insert the first argument, anObject into the collection before slot index.
     Return the receiver (sigh - ST-80 compatibility).

     Notice, that this is modifies the receiver NOT a copy.
     Also note, that it may be a slow operation for some collections,
     due to the grow:-message, which is inefficient for fixed size 
     collections (i.e. for Strings and Arrays it is not recommened)."

    |newSize|

    newSize := self size + 1.
    self grow:newSize.
    self replaceFrom:index + 1 to:newSize with:self startingAt:index.
    self at:index put:anElement

    "
     #(1 2 3 4 5 6 7 8) add:'hello' beforeIndex:5
    "
!

addFirst:anObject
    "prepend the argument, anObject to the collection.
     Return the argument, anObject.

     Notice, that this is modifies the receiver NOT a copy.
     Also note, that it may be a slow operation for some collections,
     due to the grow:-message, which is inefficient for fixed size 
     collections (i.e. for Strings and Arrays it is not recommened)."

    |newSize|

    newSize := self size + 1.
    self grow:newSize.
    self replaceFrom:2 to:newSize with:self startingAt:1.
    self at:1 put:anObject.
    ^ anObject

    "
     |a| 
     a:= #(1 2 3 4 5 6 7 8). 
     a addFirst:'hello'. 
     a 
    "
    "
     |c|
     c := #(1 2 3 4 5 6 7 8) asOrderedCollection.
     c addFirst:'hello'.
     c
    "
!

remove:anElement ifAbsent:aBlock
    "search for an object which is equal to anElement;
     if found remove and return it; if not, return the value from evaluating aBlock.
     Use equality compare (=) to search for an occurrence.

     Notice, that this is modifies the receiver NOT a copy.
     Also note, that it may be a slow operation for some collections,
     due to the grow:-message, which is inefficient for fixed size 
     collections (i.e. for Strings and Arrays it is not recommened)."

    |any 
     dstIndex "{ Class: SmallInteger }"
     sz       "{ Class: SmallInteger }"|

    dstIndex := 1.
    any := false.
    sz := self size.
    1 to:sz do:[:srcIndex |
        (anElement = (self at:srcIndex)) ifTrue:[
            any := true
        ] ifFalse:[
            (dstIndex ~~ srcIndex) ifTrue:[
                self at:dstIndex put:(self at:srcIndex)
            ].
            dstIndex := dstIndex + 1
        ]
    ].
    any ifTrue:[
        self grow:dstIndex - 1.
        ^ anElement
    ].
    ^ aBlock value

    "
     #(1 2 3 4 5 6 7 8 9 0) remove:3 ifAbsent:[Transcript showCR:'no']
     #(1 2 3 4 5 6 7 8 9 0) remove:99 ifAbsent:[#oops]
     #(1.0 2.0 3.0 4.0 5.0) remove:5 ifAbsent:[#oops]
     #(1.0 2.0 3.0 4.0 5.0) removeIdentical:5 ifAbsent:[#oops]
    "

    "Modified: 1.2.1997 / 11:49:01 / cg"
!

removeAllSuchThat:aBlock
    "remove all elements that meet a test criteria as specified in aBlock.
     The argument, aBlock is evaluated for successive elements and all those,
     for which it returns true, are removed.
     Return a collection containing the removed elements."

    "/ this is a q&d implementation (possibly slow).

    |runIndex removed element|

    removed := self species new.

    runIndex := 1.
    [runIndex <= self size] whileTrue:[
        element := self at:runIndex.
        (aBlock value:element) ifTrue:[
            removed add:element.
            self removeAtIndex:runIndex
        ] ifFalse:[
            runIndex := runIndex + 1
        ]
    ].
    ^ removed

    "
     |coll|

     coll := OrderedCollection withAll:(1 to:10).
     Transcript showCR:(coll removeAllSuchThat:[:el | el even]).
     Transcript showCR:coll
    "

    "Created: 11.5.1996 / 09:49:30 / cg"
!

removeAtIndex:anIndex
    "remove the element stored at anIndex. Return the removed object.

     Notice, that this is modifies the receiver NOT a copy.
     Also note, that it may be a slow operation for some collections,
     due to the grow:-message, which is inefficient for fixed size 
     collections (i.e. for Strings and Arrays it is not recommened)."

    |element|

    element := self at:anIndex.
    self removeFromIndex:anIndex toIndex:anIndex.
    ^ element

    "
     #(1 2 3 4 5 6 7 8 9) asOrderedCollection removeAtIndex:3
     #(1 2 3 4 5) asOrderedCollection removeAtIndex:6
     #($a $b $c $d $e $f $g) removeAtIndex:3
    "
!

removeFirst
    "remove the first element of the receiver and return it.

     Notice, that this is modifies the receiver NOT a copy.
     Also note, that it may be a slow operation for some collections,
     due to the grow:-message, which is inefficient for fixed size 
     collections (i.e. for Strings and Arrays it is not recommened)."

    ^ self removeAtIndex:1

    "
     |a|
     a := #(1 2 3 4 5 6).
     a removeFirst.
     a 
    "
!

removeFromIndex:startIndex
    "remove the elements stored at indexes from startIndex to the end.
     Return the receiver.
     Returning the receiver is a historic leftover - it may at one
     time return a collection of the removed elements.

     Notice, that this is modifies the receiver - NOT a copy; 
     therefore any other users of the receiver will also see this change.
     Also note, that it may be a slow operation for some collections,
     due to the grow:-message, which is inefficient for fixed size 
     collections (i.e. for Strings and Arrays it is not recommened)."

    ^ self removeFromIndex:startIndex toIndex:(self size)

    "
     #(1 2 3 4 5 6 7 8 9 0) asOrderedCollection removeFromIndex:3 
     #(1 2 3 4 5 6 7 8 9 0) removeFromIndex:3
    "

    "Modified: 28.1.1997 / 12:35:20 / cg"
!

removeFromIndex:startIndex toIndex:endIndex
    "remove the elements stored at indexes between startIndex and endIndex.
     Return the receiver.
     Returning the receiver is a historic leftover - it may at one
     time return a collection of the removed elements.

     Notice, that this is modifies the receiver - NOT a copy; 
     therefore any other users of the receiver will also see this change.
     Also note, that it may be a slow operation for some collections,
     due to the grow:-message, which is inefficient for fixed size 
     collections (i.e. for Strings and Arrays it is not recommened)."

    |size newSize|

    (endIndex >= (size := self size)) ifTrue:[
	self grow:(startIndex - 1)
    ] ifFalse:[
	newSize := size - endIndex + startIndex - 1.
	self replaceFrom:startIndex to:newSize with:self startingAt:(endIndex + 1).
	self grow:newSize
    ]

"/    |newSize|
"/
"/    newSize := self size - endIndex + startIndex - 1.
"/    newSize <= 0 ifTrue:[
"/        self grow:0
"/    ] ifFalse:[
"/        self replaceFrom:startIndex to:newSize with:self startingAt:(endIndex + 1).
"/        self grow:newSize
"/    ]

    "
     #(1 2 3 4 5 6 7 8 9 0) asOrderedCollection removeFromIndex:3 toIndex:5 
     #(1 2 3 4 5 6 7 8 9 0) removeFromIndex:3 toIndex:5 
     #(1 2 3 4 5 6 7 8 9 0) asOrderedCollection removeFromIndex:1 toIndex:10 
     #(1 2 3 4 5 6 7 8 9 0) removeFromIndex:1 toIndex:10 
    "
!

removeIdentical:anElement ifAbsent:aBlock
    "search for an object which is identical to anElement;
     if found remove and return it; if not, return the value from evaluating aBlock.
     Uses identity compare (==) to search for an occurrence.

     Notice, that this is modifies the receiver NOT a copy.
     Also note, that it may be a slow operation for some collections,
     due to the grow:-message, which is inefficient for fixed size 
     collections (i.e. for Strings and Arrays it is not recommened)."

    |any el
     dstIndex "{ Class: SmallInteger }"
     sz       "{ Class: SmallInteger }"|

    dstIndex := 1.
    any := false.
    sz := self size.
    1 to:sz do:[:srcIndex |
        el := self at:srcIndex.
        (anElement == el) ifTrue:[
            any := true
        ] ifFalse:[
            (dstIndex ~~ srcIndex) ifTrue:[
                self at:dstIndex put:el
            ].
            dstIndex := dstIndex + 1
        ]
    ].
    any ifTrue:[
        self grow:dstIndex - 1.
        ^ anElement
    ].
    ^ aBlock value

    "
     #(1.0 2.0 3.0 4.0 5.0) remove:5 ifAbsent:[#oops]         
     #(1.0 2.0 3.0 4.0 5.0) removeIdentical:5 ifAbsent:[#oops]
    "

    "Modified: 8.2.1997 / 16:26:49 / cg"
!

removeIndex:index
    "remove the argument stored at index. Return the receiver.

     Notice, that this is modifies the receiver NOT a copy.
     Also note, that it may be a slow operation for some collections,
     due to the grow:-message, which is inefficient for fixed size 
     collections (i.e. for Strings and Arrays it is not recommened)."

    self removeFromIndex:index toIndex:index

    "
     #(1 2 3 4 5 6 7 8 9) asOrderedCollection removeIndex:3
     #(1 2 3 4 5) asOrderedCollection removeIndex:6
     #($a $b $c $d $e $f $g) removeIndex:3
    "
!

removeLast
    "remove the last element of the receiver and return it.

     Notice, that this is modifies the receiver NOT a copy.
     Also note, that it may be a slow operation for some collections,
     due to the grow:-message, which is inefficient for fixed size
     collections (i.e. for Strings and Arrays it is not recommened)."

    ^ self removeAtIndex:(self size) 

    "
     |a|
     a := #(1 2 3 4 5 6).
     a removeLast.
     a   
    "
! !

!SequenceableCollection methodsFor:'comparing'!

= aCollection
    "return true if the receiver and aCollection represent collections
     with equal contents."

    |index "{ Class: SmallInteger }"
     stop  "{ Class: SmallInteger }" |

    (aCollection == self) ifTrue:[^true].
    (aCollection isSequenceable) ifFalse:[^false].

    stop := self size.
    stop == (aCollection size) ifFalse:[^false].

    index := 1.
    [index <= stop] whileTrue:[
	(self at:index) = (aCollection at:index) ifFalse:[^false].
	index := index + 1
    ].
    ^ true

    "
     #(1 2 3 4 5) = #(1 2 3 4 5)                        
     #($1 $2 $3 $4 $5) = #(1 2 3 4 5)                   
     #($1 $2 $3 $4 $5) = '12345'                       
     #($1 $2 $3 $4 $5) = '54321' asSortedCollection   
    "
!

endsWith:aCollection
    "return true, if the receivers last elements match those
     of aCollection"

    |index1 "{ Class: SmallInteger }"
     index2 "{ Class: SmallInteger }" 
     stop   "{ Class: SmallInteger }" 
     sz     "{ Class: SmallInteger }"|

    (aCollection == self) ifTrue:[^ true].
    (aCollection isSequenceable) ifFalse:[^ false].

    stop := aCollection size.
    sz := self size.
    stop > sz ifTrue:[^false].

    index1 := sz.
    index2 := stop.
    [index2 > 0] whileTrue:[
	(self at:index1) = (aCollection at:index2) ifFalse:[^ false].
	index1 := index1 - 1.
	index2 := index2 - 1
    ].
    ^ true

    "
     'abcde' endsWith:#($d $e)
     #[1 2 3 4] endsWith:#(3 4)    
     #(1 2 3 4) asOrderedCollection endsWith:#(3 4)
    "
!

hash
    "return a hash key for the receiver"

    "this hash is stupid - but for larger collections, the hashing
     time can become much bigger than the time lost in added probing.
     Time will show ..."

    ^ ((self at:1 ifAbsent:0) hash times:self size) bitAnd:16r3FFFFFFF

    "
     #(1 2 3 4 5) hash
     #(1 2 3 4 5.0) asOrderedCollection hash
    "
!

startsWith:aCollection
    "return true, if the receivers first elements match those
     of aCollection"

    |index "{ Class: SmallInteger }"
     stop  "{ Class: SmallInteger }" |

    (aCollection == self) ifTrue:[^true].
    (aCollection isSequenceable) ifFalse:[^false].

    stop := aCollection size.
    stop > self size ifTrue:[^false].

    index := 1.
    [index <= stop] whileTrue:[
	(self at:index) = (aCollection at:index) ifFalse:[^false].
	index := index + 1
    ].
    ^ true

    "
     'abcde' startsWith:#($a $b $c)
     #[1 2 3 4] startsWith:#(1 2 3)
     #(1 2 3 4) asOrderedCollection startsWith:#(1 2 3)
    "
! !

!SequenceableCollection methodsFor:'converting'!

asStringCollection
    "return a new string collection containing the elements;
     these ought to be strings. (i.e. String or Text instances)"

    |newColl sz|

    sz := self size.
    newColl := (StringCollection new:sz) grow:sz.
    newColl replaceFrom:1 to:sz with:self startingAt:1.
    ^ newColl

    "Created: 18.5.1996 / 13:53:55 / cg"
    "Modified: 18.5.1996 / 14:00:16 / cg"
!

asStringWith:sepChar
    "return a string generated by concatenating my elements 
     (which must be strings or nil) and embedding sepChar characters in between.
     Nil entries and empty strings are counted as empty lines."

    ^ self asStringWith:sepChar from:1 to:(self size)

    "
     #('hello' 'world' 'foo' 'bar' 'baz') asStringWith:$;
    "

    "Modified: 18.5.1996 / 15:28:14 / cg"
!

asStringWith:sepCharacter from:firstLine to:lastLine
    "return part of myself as a string with embedded sepCharacters.
     My elements must be strings or nil; nil entries and empty strings are
     taken as empty lines."

    ^ self 
        asStringWith:sepCharacter 
        from:firstLine 
        to:lastLine
        compressTabs:false
        final:nil
    "
     creating entries for searchpath:

     #('foo' 'bar' 'baz' '/foo/bar') asStringWith:$;   

     #('foo' 'bar' 'baz' '/foo/bar') asStringWith:$: from:1 to:3   
    "

    "Modified: 23.2.1996 / 15:28:55 / cg"
!

asStringWith:sepCharacter from:firstLine to:lastLine compressTabs:compressTabs final:endCharacter
    "return part of myself as a string or text with embedded sepCharacters.
     My elements must be strings or nil; nil entries and empty strings are
     taken as empty lines.
     If the argument compressTabs is true, leading spaces are converted
     to tab-characters (8col tabs). The last line is followed by a final
     character (if non-nil).

     This method is tuned for big collections, in not creating many
     intermediate strings (has linear runtime). For very small collections
     and small strings, it may be faster to use the comma , operation."

    |idx1        "{ Class:SmallInteger }"
     idx2        "{ Class:SmallInteger }"
     totalLength "{ Class:SmallInteger }"
     pos         "{ Class:SmallInteger }"
     newString lineString spaces idx nTabs sepCnt
     any16Bit stringClass needEmphasis newRuns c
     thisLen anyTab|

    "
     first accumulate the size of the string, to avoid
     countless reallocations. If tabs are compressed,
     the size computed is not exact, but gives an upper bound ...
     On the fly, look if a 16bit string or emphasized text is needed.
    "
    any16Bit := needEmphasis := false.
    stringClass := String.

    totalLength := 0.
    sepCnt := sepCharacter notNil ifTrue:[1] ifFalse:[0].

    idx1 := firstLine.
    idx2 := lastLine.
    idx1 to:idx2 do:[:lineIndex |
        lineString := self at:lineIndex.

        lineString isNil ifTrue:[
            totalLength := totalLength + sepCnt
        ] ifFalse: [
            lineString hasChangeOfEmphasis ifTrue:[
                needEmphasis := true
            ].
            (lineString bitsPerCharacter == 16) ifTrue:[
                any16Bit := true.
                stringClass := lineString class
            ].
            totalLength := totalLength + lineString size + sepCnt
        ].
    ].
    endCharacter isNil ifTrue:[
        totalLength := totalLength - 1.
        totalLength < 0 ifTrue:[^ ''].
    ].
    spaces := '        '.
    newString := stringClass new:totalLength.

    needEmphasis ifTrue:[
        newRuns := RunArray new.
    ].

    "
     now, replace ...
     Be careful with runArrays:
        replacing individual elements is VERY expensive.
        Therefore, create a new runArray from scratch.
    "
    pos := 1.
    idx1 to:idx2 do:[:lineIndex |
        lineString := self at:lineIndex.
        thisLen := lineString size.
        thisLen ~~ 0 ifTrue:[
            (anyTab := compressTabs) ifTrue:[
                "
                 mhmh: could use withTabs from String-class here,
                 but we should avoid creating too many temporary strings
                 (especially, since this method is typically used when converting
                 big texts such as when saving in the filebrowser ...).
                 Therefore, we convert tabs inline here doing a direct replace
                 in newString."

                idx := lineString findFirst:[:c | (c ~~ Character space)].
                nTabs := (idx-1) // 8.
                nTabs ~~ 0 ifTrue:[
                    anyTab := true
                ]
            ].
            anyTab ifTrue:[
                "any tabs"

                idx := nTabs * 8 + 1.   "/ index of first copied character in string

                newString atAll:(pos to:pos+nTabs-1) put:(Character tab).
                newRuns notNil ifTrue:[
                    newRuns add:nil withOccurrences:nTabs
                ].
                pos := pos + nTabs.

                newString replaceFrom:pos with:lineString startingAt:idx.
                newRuns notNil ifTrue:[
                    lineString hasChangeOfEmphasis ifTrue:[
                        idx to:lineString size do:[:pos |
                            newRuns add:(lineString emphasisAt:pos)
                        ]    
                    ] ifFalse:[
                        newRuns add:nil withOccurrences:(lineString size - idx + 1)
                    ]
                ].
                pos := pos + thisLen - (nTabs * 8).
            ] ifFalse:[
                newString replaceFrom:pos with:lineString.

                newRuns notNil ifTrue:[
                    lineString hasChangeOfEmphasis ifTrue:[
                        newRuns addAll:(lineString emphasis)
                    ] ifFalse:[
                        newRuns add:nil withOccurrences:lineString size
                    ]
                ].
                pos := pos + thisLen.
            ].
        ].

        lineIndex ~~ lastLine ifTrue:[
            c := sepCharacter
        ] ifFalse:[
            c := endCharacter
        ].

        c notNil ifTrue:[
            newString at:pos put:c.
            newRuns notNil ifTrue:[
                newRuns add:nil.
            ].
            pos := pos + 1
        ].
    ].

    "
     in case of tab compression, the result has to be
     cut to size ... sorry
    "
    pos ~~ totalLength ifTrue:[
        newString := newString copyTo:(pos - 1)
    ].

    newRuns notNil ifTrue:[
        newString := Text string:newString runs:newRuns.
    ].

    ^ newString

    "
     creating entries for searchpath:

       #('foo' 'bar' 'baz' '/foo/bar') 
          asStringWith:$: 
          from:1 to:4 
          compressTabs:false 
          final:nil 

     with trailing colon:

       #('foo' 'bar' 'baz' '/foo/bar') 
          asStringWith:$: 
          from:1 to:4 
          compressTabs:false 
          final:$: 

     concatenating all elements:

       #('foo' 'bar' 'baz') 
          asStringWith:nil 
          from:1 to:3 
          compressTabs:false 
          final:nil 

     creating a string from a collection of lines:

       #('foo' 'bar' 'baz') 
          asStringWith:(Character cr) 
          from:1 to:3 
          compressTabs:false 
          final:(Character cr) 

     creating a text from a collection of mixed texts and strings:

       (Array
            with:'foo'
            with:('bar' asText allBold)
            with:'baz'
            with:('baz2' asText emphasizeAllWith:#italic)
       ) 
          asStringWith:(Character cr) 
          from:1 to:4 
          compressTabs:false 
          final:(Character cr) 
    "

    "Modified: 16.8.1997 / 04:44:46 / cg"
!

asStringWithCRs
    "return a string generated by concatenating my elements 
     (which must be strings or nil) and embedding cr characters in between.
     Nil entries and empty strings are counted as empty lines."

    ^ self asStringWithCRsFrom:1 to:(self size)

    "
     #('hello' 'world' 'foo' 'bar' 'baz') asStringWithCRs 

     (OrderedCollection new
        add:'hello'; 
        add:'world';
        add:'foo';
        add:('bar' asText allBold);
        yourself) asStringWithCRs

     Transcript showCR:
         (OrderedCollection new
            add:'hello'; 
            add:'world';
            add:'foo';
            add:('bar' asText allBold);
            yourself) asStringWithCRs
    "

    "Modified: 18.5.1996 / 16:43:47 / cg"
!

asStringWithCRsFrom:firstLine to:lastLine
    "return a string generated by concatenating some of my elements 
     (which must be strings or nil) and embedding cr characters in between.
     Nil entries and empty strings are counted as empty lines."

    ^ self asStringWithCRsFrom:firstLine to:lastLine compressTabs:false withCR:true

    "
     #('hello' 'world' 'foo' 'bar' 'baz') asStringWithCRsFrom:2 to:4 

    "

    "Modified: 18.5.1996 / 16:50:55 / cg"
!

asStringWithCRsFrom:firstLine to:lastLine compressTabs:compressTabs
    "return part of myself as a string with embedded cr's.
     My elements must be strings or nil.
     If the argument compressTabs is true, leading spaces are converted
     to tab-characters (8col tabs).
     Nil entries and empty strings are taken as empty lines."

    ^ self asStringWithCRsFrom:firstLine to:lastLine compressTabs:compressTabs withCR:true
!

asStringWithCRsFrom:firstLine to:lastLine compressTabs:compressTabs withCR:withCR
    "return part of myself as a string with embedded cr's.
     My elements must be strings or nil; nil entries and empty strings are
     taken as empty lines.
     If the argument compressTabs is true, leading spaces are converted
     to tab-characters (8col tabs). WithCR controls whether the last line
     should be followed by a cr or not."

    ^ self asStringWith:(Character cr)
		   from:firstLine 
		     to:lastLine 
	   compressTabs:compressTabs 
		  final:(withCR ifTrue:[Character cr] ifFalse:[nil])
!

decodeAsLiteralArray
    "given a literalEncoding in the receiver,
     create & return the corresponding object.
     The inverse operation to #literalArrayEncoding."

    |clsName cls|

    clsName := self first.
    cls := Smalltalk at:clsName ifAbsent:nil.
    cls isNil ifTrue:[
        ^ self error:'unknown class in spec: ' , clsName.
    ].
    ^ cls decodeFromLiteralArray:self.

    "
     #(Point 10 20) decodeAsLiteralArray  
     #(Rectangle 100 200 400 500) decodeAsLiteralArray 
     #(#MenuItem #rawLabel: 'right' #value: #right) decodeAsLiteralArray 
    "

    "Modified: 25.2.1997 / 19:52:40 / cg"
!

literalArrayEncoding
    "encode myself as an array literal, from which a copy of the receiver
     can be reconstructed with #decodeAsLiteralArray."

    ^ (self collect:[:el | el literalArrayEncoding]) asArray

    "
     (Array with:(Color red:50 green:50 blue:50)
            with:(1 @ 2)
     ) literalArrayEncoding   
    "

    "Modified: 22.4.1996 / 13:00:56 / cg"
! !

!SequenceableCollection methodsFor:'copying'!

, aCollection
    "return a new collection formed from concatenating the receiver with
     the argument. The class of the new collection is determined by the
     receivers class, so mixing classes is possible, if the second collections
     elements can be stored into instances of the receivers class."

    |newCollection 
     mySize    "{ Class: SmallInteger }"
     newSize   "{ Class: SmallInteger }"
     otherSize "{ Class: SmallInteger }"
     dstIndex  "{ Class: SmallInteger }"|

    mySize := self size.
    otherSize := aCollection size.
    newSize := mySize + otherSize.

    newCollection := self copyEmptyAndGrow:newSize.   "must grow, otherwise replace fails"

    newCollection replaceFrom:1 to:mySize with:self startingAt:1.
    dstIndex := mySize + 1.
    aCollection isSequenceable ifTrue:[
	"
	 yes, aCollection has indexed elements
	"
	newCollection replaceFrom:dstIndex to:newSize
			     with:aCollection startingAt:1.
	^ newCollection
    ] ifFalse:[
	"
	 no, enumerate aCollection
	"
	aCollection do:[:element |
	    newCollection at:dstIndex put:element.
	    dstIndex := dstIndex + 1
	]
    ].
    ^ newCollection

    "
     #($a $b $c) , #(1 2 3)
     #($a $b $c) , '123'
     'abc' , '123'
     'abc' , #($q $w $e $r $t $y) asSortedCollection
     'abc' , #(1 2 3 4 5)"  "-- will fail, since strings cannot store integers
     'abc' asArray , #(1 2 3 4 5)
    "
!

copyFirst:count
    "return a new collection consisting of the receivers first count
     elements - this is just a rename of copyTo: - for compatibility."

    ^ self copyFrom:1 to:count

    "
     #($a $b $c $d $e $f $g) copyFirst:5
     '1234567890' copyFirst:4
    "
!

copyFrom:start
    "return a new collection consisting of receivers elements
     from start to the end of the collection"

    ^ self copyFrom:start to:(self size)

    "
     #($a $b $c $d $e $f $g) copyFrom:2
     '1234567890' copyFrom:2
    "
!

copyFrom:start count:numberOfElements
    "return a new collection consisting of numberOfElements elements
     extracted from the receiver starting at index start
     (i.e. extract a slice)"

    ^ self
        copyFrom:start
        to:(start + numberOfElements - 1)

    "
     #($a $b $c $d $e $f $g) copyFrom:2 count:3 
     '1234567890' copyFrom:2 count:5    
    "

    "Modified: 20.2.1997 / 14:23:01 / cg"
!

copyFrom:start to:stop
    "return a new collection consisting of receivers elements
     between start and stop"

    |newCollection newSize|

    newSize := stop - start + 1.
    newCollection := self copyEmptyAndGrow:newSize.
    newCollection replaceFrom:1 to:newSize with:self startingAt:start.
    ^ newCollection

    "
     #($a $b $c $d $e $f $g) copyFrom:2 to:5
     '1234567890' copyFrom:2 to:5
    "
!

copyLast:count
    "return a new collection consisting of the receivers last count
     elements."

    |sz|

    sz := self size.
    ^ self copyFrom:(sz - count + 1) to:sz

    "
     #($a $b $c $d $e $f $g) copyLast:5
     '1234567890' copyLast:4
    "
!

copyReplaceFrom:startIndex to:endIndex with:aCollection
    "return a copy of the receiver, where the elements from startIndex to
     endIndex have been replaced by the elements of aCollection"

    |newColl sz replSize|

    replSize := aCollection size.
    sz := self size - (endIndex - startIndex + 1) + replSize.
    newColl := self copyEmptyAndGrow:sz.
    newColl replaceFrom:1 to:(startIndex - 1) with:self.
    newColl replaceFrom:startIndex with:aCollection.
    newColl replaceFrom:(startIndex + replSize) with:self startingAt:(endIndex + 1).
    ^ newColl

    "
     #(1 2 3 4 5 6 7 8 9 0) copyReplaceFrom:3 to:6 with:#(a b c d e f g h i j k)
     'hello world' copyReplaceFrom:6 to:6 with:' there, '  
    "
!

copyReplaceFrom:startIndex with:aCollection
    "return a copy of the receiver, where the elements from startIndex to
     the end have been replaced by the elements of aCollection"

    ^ self copyReplaceFrom:startIndex to:(self size) with:aCollection

    "
     #(1 2 3 4 5 6 7 8 9 0) copyReplaceFrom:3 with:#(a b c)  
     'hello world' copyReplaceFrom:7 with:'smalltalk fan'  
    "
!

copyThrough:element
    "return a new collection consisting of the receiver elements
     up-to (AND including) the first occurence of element."

    |idx|

    idx := self indexOf:element.
    idx == 0 ifTrue:[^ nil].    "question: is this ok?"
    ^ self copyFrom:1 to:idx

    "
     #($a $b $c $d $e $f $g) copyThrough:$d
     '1234567890' copyThrough:$5
     '1234567890' copyThrough:$a
     '1234567890' copyThrough:$1
    "
!

copyTo:stop
    "return a new collection consisting of receivers elements
     from 1 up to (including) index stop"

    ^ self copyFrom:1 to:stop

    "
     #($a $b $c $d $e $f $g) copyTo:5
     '1234567890' copyTo:4
    "
!

copyUpTo:element
    "return a new collection consisting of the receiver elements
     up-to (but excluding) the first occurence of element;
     or to the end, if element is not included"

    |idx|

    idx := self indexOf:element.
    idx == 0 ifTrue:[^ self copy].    "question: is this ok?"
    idx == 1 ifTrue:[^ self copyEmpty].
    ^ self copyFrom:1 to:(idx-1)

    "
     #($a $b $c $d $e $f $g) copyUpTo:$d
     '1234567890' copyUpTo:$5
     '1234567890' copyUpTo:$a
     '1234567890' copyUpTo:$1
     '123456789' copyUpTo:$0
    "
!

copyWith:newElement
    "return a new collection containing the receivers elements
     and the single new element, newElement. 
     This is different from concatentation, which expects another collection
     as argument, but equivalent to copy-and-addLast."

    |newCollection mySize newSize|

    mySize := self size.
    newSize := mySize + 1.
    newCollection := self copyEmptyAndGrow:newSize.
    newCollection replaceFrom:1 to:mySize with:self startingAt:1.
    newCollection at:newSize put:newElement.
    ^newCollection

    "
     #(1 2 3 4 5) copyWith:$a
     'abcdefg' copyWith:$h
     'abcdefg' copyWith:'123'   -- will fail: string cannot be stored into string
     'abcdefg' copyWith:1       -- will fail: integer cannot be stored into string
    "
!

copyWithout:elementToSkip
    "return a new collection consisting of a copy of the receiver, with
     ALL elements equal to elementToSkip are left out.
     No error is reported, if elementToSkip is not in the collection."

    |n         "{ Class: SmallInteger }"
     sz        "{ Class: SmallInteger }"
     srcIndex  "{ Class: SmallInteger }"
     dstIndex  "{ Class: SmallInteger }"
     skipIndex "{ Class: SmallInteger }"
     copy l|

    "the code below may look like overkill, 
     however, for big collections its better to move data
     around in big chunks"

    n := self occurrencesOf:elementToSkip.
    n == 0 ifTrue:[^ self copy].

    sz := self size.
    copy := self copyEmptyAndGrow:(sz - n).

    srcIndex := 1.
    dstIndex := 1.

    n timesRepeat:[
	skipIndex := self indexOf:elementToSkip startingAt:srcIndex.
	l := skipIndex - srcIndex.
	l ~~ 0 ifTrue:[
	    copy replaceFrom:dstIndex to:(dstIndex + l - 1) 
			with:self startingAt:srcIndex.
	    dstIndex := dstIndex + l
	].
	srcIndex := skipIndex + 1
    ].
    l := sz - srcIndex.
    copy replaceFrom:dstIndex to:(dstIndex + l)
		with:self startingAt:srcIndex.
    ^ copy

    "
     #($a $b $c $d $e $f $g) copyWithout:$d
     #($a $b $c $d $e $f $g) copyWithout:$a
     #($a $b $c $d $e $f $g) copyWithout:$g
     'abcdefghi' copyWithout:$h    
     'abcdefg' copyWithout:$h       
     #($a $b $c $a $a $d $e $a $f $g) copyWithout:$a
     #($a $b $c $d $e $f $g) copyWithout:$x
     #(90 80 70 60 50) copyWithout:70
     #(90 80 70 80 60 45 80 50) copyWithout:80
    "
!

copyWithoutFirst:elementToSkip
    "return a new collection consisting of a copy of the receivers elements
     without the first elementToSkip, if it was present. 
     No error is reported, if elementToSkip is not in the collection."

    |copy skipIndex sz|

    skipIndex := self indexOf:elementToSkip startingAt:1.
    (skipIndex == 0) ifTrue:[^ self copy].

    sz := self size - 1.
    copy := self copyEmptyAndGrow:sz.
    copy replaceFrom:1 to:(skipIndex - 1) with:self startingAt:1.
    copy replaceFrom:skipIndex to:sz with:self startingAt:(skipIndex + 1).
    ^ copy

    "
     #($a $b $c $d $e $f $g) copyWithoutFirst:$d
     #($a $b $c $d $e $f $g) copyWithoutFirst:$x
     #(90 80 70 60 50) copyWithoutFirst:70
     #(90 80 70 80 60 45 80 50) copyWithoutFirst:80
    "
!

copyWithoutIndex:omitIndex
    "return a new collection consisting of receivers elements
     without the argument stored at omitIndex"

    |copy sz|

    sz := self size - 1.
    copy := self copyEmptyAndGrow:sz.
    copy replaceFrom:1 to:(omitIndex - 1) with:self startingAt:1.
    copy replaceFrom:omitIndex to:sz with:self startingAt:(omitIndex + 1).
    ^ copy

    "
     #(1 2 3 4 5 6 7 8 9 0) copyWithoutIndex:3
     'abcdefghijkl' copyWithoutIndex:5
    "
!

copyWithoutLast:count
    "return a new collection consisting of the receivers elements
     except the last count elements."

    |sz|

    sz := self size.
    ^ self copyTo:(sz - count)

    "
     #($a $b $c $d $e $f $g) copyWithoutLast:5 
     '1234567890' copyWithoutLast:4 
    "
! !

!SequenceableCollection methodsFor:'enumerating'!

collect:aBlock
    "evaluate the argument, aBlock for every element in the collection
     and return a collection of the results"

    |newCollection
     sz  "{ Class:SmallInteger }"|

    sz := self size.
    newCollection := self copyEmptyAndGrow:sz.
    1 to:sz do:[:index |
	newCollection at:index put:(aBlock value:(self at:index)).
    ].
    ^ newCollection

    "
     #(one two three four five six) collect:[:element | element asUppercase]  
     #(1 2 3 4 5 6 7 8 9) collect:[:element | element factorial]   
     (1 to:9) collect:[:element | element * element]   
    "
!

do:aBlock
    "evaluate the argument, aBlock for every element in the collection."

    |stop "{ Class:SmallInteger }"|

    stop := self size.
    1 to:stop do:[:index |
	aBlock value:(self at:index).
    ]

    "
     #(one two three four five six) do:[:element | Transcript showCR:element]
    "
!

do:aBlock separatedBy:sepBlock
    "evaluate the argument, aBlock for every element in the collection.
     Between each element (i.e. not before the first element and not after the
     last element), evaluate sepBlock.
     This supports printing of collections with elements separated by some
     string."

    |stop "{ Class:SmallInteger }"|

    stop := self size.
    stop < 1 ifTrue:[^ self].
    aBlock value:(self at:1).
    2 to:stop do:[:index |
	sepBlock value.
	aBlock value:(self at:index)
    ].

    "
     #(one two three four five six)
	do:[:each | Transcript show:each]
	separatedBy:[Transcript show:' , ']
    "
!

doWithIndex:aBlock
    "Squeak compatibility; like keysAndValuesDo:, but passes
     the index as second argument."

    self keysAndValuesDo:[:index :value |
        aBlock value:value value:index
    ].

    "Created: 17.10.1997 / 12:33:10 / cg"
!

from:startIndex do:aBlock
    "evaluate the argument, aBlock for the elements starting with the
     element at startIndex to the end."

    ^ self from:startIndex to:self size do:aBlock

    "
     #(one two three four five six) 
	from:3 
	do:[:element | Transcript showCR:element]
    "
!

from:start to:stop collect:aBlock
    "evaluate the argument, aBlock for the elements indexed by start
     to stop in the collection and return a collection of the results"

    |newCollection sz
     idx  "{ Class:SmallInteger }"|

    sz := stop - start + 1.
    newCollection := self copyEmptyAndGrow:sz.
    idx := 1.
    start to:stop do:[:index |
	newCollection at:idx put:(aBlock value:(self at:index)).
	idx := idx + 1
    ].
    ^ newCollection

    "
     #(one two three four five six) 
	from:2 
	to:4 
	collect:[:element | element asUppercase]  
    "
!

from:index1 to:index2 do:aBlock
    "evaluate the argument, aBlock for the elements with index index1 to
     index2 in the collection"

    |start "{ Class:SmallInteger }"
     stop  "{ Class:SmallInteger }" |

    start := index1.
    stop := index2.
    start to:stop do:[:index |
	aBlock value:(self at:index).
    ]

    "
     #(one two three four five six) 
	from:3 
	to:5 
	do:[:element | Transcript showCR:element]
    "
!

from:index1 to:index2 reverseDo:aBlock
    "evaluate the argument, aBlock for the elements with index index1 to
     index2 in the collection. Step in reverse order"

    |start "{ Class:SmallInteger }"
     stop  "{ Class:SmallInteger }" |

    start := index1.
    stop := index2.
    stop to:start by:-1 do:[:index |
	aBlock value:(self at:index).
    ]

    "
     #(one two three four five six) 
	from:3 
	to:5 
	reverseDo:[:element | Transcript showCR:element]
    "
!

keysAndValuesDo:aTwoArgBlock
    "evaluate the argument, aBlock for every element in the collection,
     passing both index and element as arguments."

    |stop  "{ Class:SmallInteger }"|

    stop := self size.
    1 to:stop do:[:index |
	aTwoArgBlock value:index value:(self at:index).
    ]
    "
     #(one two three four five six) 
	keysAndValuesDo:[:key :element | 
			    Transcript show:key; space; showCR:element
			]
    "
!

keysAndValuesReverseDo:aTwoArgBlock
    "evaluate the argument, aBlock in reverse order, for every element 
     in the collection, passing both index and element as arguments."

    |stop  "{ Class:SmallInteger }"|

    stop := self size.
    stop to:1 by:-1 do:[:index |
        aTwoArgBlock value:index value:(self at:index).
    ]

    "
     #(one two three four five six) 
        keysAndValuesReverseDo:[:key :element | 
				    Transcript show:key; space; showCR:element
			       ]
    "

    "Modified: 9.5.1996 / 00:57:23 / cg"
!

nonNilElementsDo:aBlock
    "evaluate the argument, aBlock for every non-nil element in the collection."

    |stop "{ Class:SmallInteger }"
     element|

    stop := self size.
    1 to:stop do:[:index |
	(element := self at:index) notNil ifTrue:[
	    aBlock value:element.
	]
    ]
    "
     #(one nil three nil five nil seven) nonNilElementsDo:[:element | Transcript showCR:element]
    "
!

pairWiseCollect:aBlock
    "evaluate the argument, aBlock for every pair of elements in the collection.
     The block is called with 2 arguments for each 2 elements in the receiver.
     An error will be reported, if the number of elements in the receiver
     is not a multiple of 2.
     Collect the results and return a new collection containing those."

    |stop newCollection dstIdx "{ Class:SmallInteger }"|

    stop := self size.
    newCollection := self copyEmptyAndGrow:stop // 2.
    dstIdx := 1.
    1 to:stop by:2 do:[:index |
	newCollection at:dstIdx put:(aBlock value:(self at:index) value:(self at:index+1)).
	dstIdx := dstIdx + 1
    ].
    ^ newCollection

    "
     #(1 one 2 two 3 three 4 four 5 five 6 six) 
     pairWiseCollect:[:num :sym | sym->num] 


     #(1 1  1 2  1 3  1 4  1 5) 
     pairWiseCollect:[:x :y | x@y] 
    "
!

pairWiseDo:aBlock
    "evaluate the argument, aBlock for every pair of elements in the collection.
     The block is called with 2 arguments for each 2 elements in the receiver.
     An error will be reported, if the number of elements in the receiver
     is not a multiple of 2."

    |stop "{ Class:SmallInteger }"|

    stop := self size.
    1 to:stop by:2 do:[:index |
	aBlock value:(self at:index) value:(self at:index+1).
    ]
    "
     #(1 one 2 two 3 three 4 four 5 five 6 six) 
     pairWiseDo:[:num :sym | Transcript show:num; show:' is: '; showCR:sym]


     #(1 1  1 2  1 3  1 4  1 5) 
     pairWiseDo:[:x :y | Transcript showCR:x@y]
    "
!

reverseDo:aBlock
    "evaluate the argument, aBlock for every element in the collection
     in reverse order"

    |sz  "{ Class:SmallInteger }"|

    sz := self size.
    sz to:1 by:-1 do:[:index |
	aBlock value:(self at:index).
    ]

    "
     #(one two three four five six) reverseDo:[:element | Transcript showCR:element]
    "
!

select:aBlock
    "evaluate the argument, aBlock for every element in the collection
     and return a collection of all elements for which the block return
     true"

    |element newColl species needCopy
     sz  "{ Class:SmallInteger }"|

    sz := self size.
    species := self species.
    species growIsCheap ifFalse:[
	newColl := OrderedCollection new:sz.
	needCopy := true
    ] ifTrue:[
	newColl := self copyEmpty:sz.
	needCopy := false
    ].
    1 to:sz do:[:index |
	element := self at:index.
	(aBlock value:element) ifTrue:[
	    newColl add:element
	].
    ].
    needCopy ifTrue:[
	newColl := (species withAll:newColl) postCopyFrom:self
    ].
    ^ newColl

    "
     #(one two three four five six) select:[:element | element startsWith:'f']   
     #(1 2 3 4 5 6 7 8 9) select:[:element | element odd]   
     (#(17 12 1 98 51) asSortedCollection:[:a :b | b < a]) select:[:element | element odd]   
     (1 to:9) select:[:element | element odd]   
     (Smalltalk allClasses) select:[:class | class name startsWith:'S']   
    "
!

with:aSequenceableCollection do:aTwoArgBlock
    "evaluate the argument, aBlock for successive elements from
     each the receiver and the argument, aSequenceableCollection.
     The second argument, aBlock must be a two-argument block.
     The collection argument must implement access via a numeric key."

    |stop  "{ Class: SmallInteger }" |

    stop := self size.
    1 to:stop do:[:index |
	aTwoArgBlock value:(self at:index) value:(aSequenceableCollection at:index).
    ]
    "
     #(one two three four five six) with:(1 to:10) do:[:el1 :el2 | Transcript show:el1; space; showCR:el2]
    "
! !

!SequenceableCollection methodsFor:'filling and replacing'!

atAllPut:anObject
    "replace all elements of the collection by the argument, anObject.
     Notice: This operation modifies the receiver, NOT a copy;
     therefore the change may affect all others referencing the receiver."

    self from:1 to:(self size) put:anObject

    "
     (Array new:10) atAllPut:1
     (String new:10) atAllPut:$a
    "
!

from:index1 to:index2 put:anObject
    "replace the elements from index1 to index2 of the collection
     by the argument, anObject.
     Notice: This operation modifies the receiver, NOT a copy;
     therefore the change may affect all others referencing the receiver."

    |index "{ Class: SmallInteger }"
     end   "{ Class: SmallInteger }"|

    index := index1.
    end := index2.
    [index <= end] whileTrue:[
	self at:index put:anObject.
	index := index + 1
    ]

    "
     #(1 2 3 4 5 6 7 8 9 0) from:3 to:6 put:$X
     'abcdefghijkl' from:3 to:6 put:$X
    "
!

replaceAll:oldObject by:newObject
    "backward ST/X compatibility: an alias for #replaceAll:with:.
     Pleace do no longer use this one;
     instead use #replaceAll:with: for ST-80 compatibility."

"/    self obsoleteMethodWarning:'use #replaceAll:with:'.
    ^ self replaceAll:oldObject with:newObject

    "Modified: 23.9.1997 / 02:17:49 / cg"
!

replaceAll:oldObject by:newObject from:startIndex to:stopIndex
    "backward ST/X compatibility: an alias for #replaceAll:with:from:to:.
     Pleace do no longer use this one;
     instead use #replaceAll:with:from:to: for ST-80 compatibility."

    ^ self replaceAll:oldObject with:newObject from:startIndex to:stopIndex
!

replaceAll:oldObject with:newObject
    "replace all oldObjects by newObject in the receiver.
     Notice: This operation modifies the receiver, NOT a copy;
     therefore the change may affect all others referencing the receiver."

    ^ self replaceAll:oldObject with:newObject from:1 to:self size

    "
     '123123abc123' replaceAll:$1 with:$*  
     #(1 2 3 4 1 2 3 4) copy replaceAll:1 with:'one' 
    "
!

replaceAll:oldObject with:newObject from:startIndex to:stopIndex
    "replace all oldObjects found between startIndex and endIndex,
     by newObject in the receiver.
     Notice: This operation modifies the receiver, NOT a copy;
     therefore the change may affect all others referencing the receiver."

    |start "{ Class: SmallInteger }"
     stop  "{ Class: SmallInteger }"|

    start := startIndex.
    stop := stopIndex.
    startIndex to:stopIndex do:[:index |
	(self at:index) = oldObject ifTrue:[
	    self at:index put:newObject
	]
    ]

    "
     '123123abc123' replaceAll:$1 with:$*
     '123123abc123' replaceAll:$1 with:$* from:1 to:6
     #(1 2 3 4 1 2 3 4) replaceAll:1 with:'one' from:1 to:4
    "
!

replaceAny:aCollection by:newObject
    "backward ST/X compatibility; alias for #replaceAny:with:.
     Pleace do no longer use this one;
     instead use #replaceAny:with: for naming consistence."

"/    self obsoleteMethodWarning:'use #replaceAny:with:'.
    ^ self replaceAny:aCollection with:newObject from:1 to:self size 

    "
     '123123abc123' replaceAny:#($1 $2) with:$*
     #('foo' 'bar' 'foo' 'baz' foo 1 2 3) replaceAny:#(foo 1) with:'*'
    "

    "Modified: 23.9.1997 / 02:17:53 / cg"
!

replaceAny:aCollection by:newObject from:startIndex to:stopIndex
    "backward ST/X compatibility; alias for #replaceAny:with:from:to:.
     Pleace do no longer use this one;
     instead use #replaceAny:with:from:to: for naming consistence."
    
"/    self obsoleteMethodWarning:'use #replaceAny:with:from:to:'.
    ^ self replaceAny:aCollection with:newObject from:startIndex to:stopIndex 

    "
     '123123abc123' replaceAny:#($1 $2) with:$* from:1 to:6      
     #('foo' 'bar' 'foo' 'baz' foo 1 2 3) replaceAny:#(foo 1) with:'*'  
    "

    "Modified: 23.9.1997 / 02:17:58 / cg"
!

replaceAny:aCollection with:newObject
    "replace all elements, which are in aCollection, by newObject.
     Notice: This operation modifies the receiver, NOT a copy;
     therefore the change may affect all others referencing the receiver."

    ^ self replaceAny:aCollection with:newObject from:1 to:self size
!

replaceAny:aCollection with:newObject from:startIndex to:stopIndex
    "replace all elements within a range,
     which are in contained in aCollection by newObject.
     Notice: This operation modifies the receiver, NOT a copy;
     therefore the change may affect all others referencing the receiver."

    |start "{ Class: SmallInteger }"
     stop "{ Class: SmallInteger }"|

    start := startIndex.
    stop := stopIndex.
    startIndex to:stopIndex do:[:index |
	(aCollection includes:(self at:index)) ifTrue:[
	    self at:index put:newObject
	]
    ]

    "
     '123123abc123' replaceAny:#($1 $2) with:$* from:1 to:6      
     #('foo' 'bar' 'foo' 'baz' foo 1 2 3) replaceAny:#(foo 1) with:'*'  
    "
!

replaceFrom:start count:numberOfElements with:replacementCollection
    "replace numberOfElements elements in the receiver from index start,
     with elements taken from replacementCollection starting at 1.
     Notice: This operation modifies the receiver, NOT a copy;
     therefore the change may affect all others referencing the receiver."

    ^ self replaceFrom:start
                    to:(start + numberOfElements - 1)
                  with:replacementCollection
            startingAt:1

    "
     '1234567890' replaceFrom:5 count:3 with:'abcdef'  
     #($a $b $c $d $e) replaceFrom:2 count:2 with:'12345'
    "

    "Modified: 20.2.1997 / 14:24:38 / cg"
!

replaceFrom:start count:numberOfElementsToReplace with:replacementCollection startingAt:repStart
    "replace numberOfElementsToReplace elements in the receiver from index start,
     with elements  taken from replacementCollection starting at repStart.
     Notice: This operation modifies the receiver, NOT a copy;
     therefore the change may affect all others referencing the receiver."

    ^ self
        replaceFrom:start
        to:(start + numberOfElementsToReplace - 1)
        with:replacementCollection
        startingAt:repStart

    "
     '1234567890' replaceFrom:5 count:3 with:'abcdef' startingAt:3  
     #($a $b $c $d $e) replaceFrom:2 count:3 with:'12345' startingAt:4 
     #($a $b $c $d $e) asOrderedCollection replaceFrom:2 count:3 with:'12345' startingAt:4 
    "

    "Modified: 20.2.1997 / 14:26:12 / cg"
!

replaceFrom:start to:stop with:replacementCollection
    "replace elements in the receiver between index start and stop,
     with elements taken from replacementCollection starting at 1.
     Notice: This operation modifies the receiver, NOT a copy;
     therefore the change may affect all others referencing the receiver."

    ^ self replaceFrom:start
		    to:stop
		  with:replacementCollection
	    startingAt:1

    "
     '1234567890' replaceFrom:5 to:7 with:'abcdef'
     #($a $b $c $d $e) replaceFrom:2 to:3 with:'12345'
    "
!

replaceFrom:start to:stop with:replacementCollection startingAt:repStart
    "replace elements in the receiver between index start and stop,
     with elements  taken from replacementCollection starting at repStart.
     Notice: This operation modifies the receiver, NOT a copy;
     therefore the change may affect all others referencing the receiver."

    |srcIndex "{ Class: SmallInteger }"
     dstIndex "{ Class: SmallInteger }"
     end      "{ Class: SmallInteger }" |

    (repStart < start) ifTrue:[
	" must do reverse copy "
	srcIndex := repStart + (stop - start).
	dstIndex := stop.
	end := start.
	[dstIndex >= end] whileTrue:[
	    self at:dstIndex put:(replacementCollection at:srcIndex).
	    srcIndex := srcIndex - 1.
	    dstIndex := dstIndex - 1
	].
	^ self
    ].

    srcIndex := repStart.
    dstIndex := start.
    end := stop.
    [dstIndex <= end] whileTrue:[
	self at:dstIndex put:(replacementCollection at:srcIndex).
	srcIndex := srcIndex + 1.
	dstIndex := dstIndex + 1
    ]

    "
     '1234567890' replaceFrom:5 to:7 with:'abcdef' startingAt:3  
     #($a $b $c $d $e) replaceFrom:2 to:3 with:'12345' startingAt:4 
     #($a $b $c $d $e) asOrderedCollection replaceFrom:2 to:3 with:'12345' startingAt:4 

     |c|
     c := #($a $b $c $d $e) asOrderedCollection.
     c replaceFrom:2 to:3 with:c startingAt:4  

     |c|
     c := #($a $b $c $d $e) asOrderedCollection.
     c replaceFrom:4 to:5 with:c startingAt:2  
    "
!

replaceFrom:start with:replacementCollection
    "replace elements in the receiver starting at start,
     with elements taken from replacementCollection starting at 1
     to the end of replacementCollection.
     Notice: This operation modifies the receiver, NOT a copy;
     therefore the change may affect all others referencing the receiver."

    ^ self replaceFrom:start 
		    to:(start + replacementCollection size - 1)
		  with:replacementCollection
	    startingAt:1

    "
     '1234567890' replaceFrom:5 with:'abc'
     #($a $b $c $d $e) replaceFrom:2 with:'123'
    "
!

replaceFrom:start with:replacementCollection startingAt:offset
    "replace elements in the receiver starting at start,
     with elements taken from replacementCollection starting at offset
     to the end of replacementCollection.
     Notice: This operation modifies the receiver, NOT a copy;
     therefore the change may affect all others referencing the receiver."

    ^ self replaceFrom:start 
		    to:(start + replacementCollection size - offset)
		  with:replacementCollection
	    startingAt:offset

    "
     '1234567890' replaceFrom:5 with:'abcdef' startingAt:3
     #($a $b $c $d $e) replaceFrom:2 with:'12345' startingAt:4
    "
!

startingAt:sourceStart replaceElementsIn:destColl from:destStartIndex to:destEndIndex
    "replace elements in destColl with elements from the receiver.
     Notice: This operation modifies the destination collection, NOT a copy;
     therefore the change may affect all others referencing this object."

    destColl replaceFrom:destStartIndex to:destEndIndex with:self startingAt:sourceStart

    "
     |s|
     s := 'abcdefghijklmnop'.
     '1234567890' startingAt:1 replaceElementsIn:s from:1 to:3.
     s'123defghijklmnop'
    "
! !

!SequenceableCollection methodsFor:'private sorting helpers'!

quickSortFrom:inBegin to:inEnd
    "actual quicksort worker for sort-message"

    |begin   "{ Class: SmallInteger }"
     end     "{ Class: SmallInteger }"
     b       "{ Class: SmallInteger }"
     e       "{ Class: SmallInteger }"
     middleElement temp1 temp2 |

    begin := inBegin.   "/ this also does a type-check
    end := inEnd.

    b := begin.
    e := end.
    middleElement := self at:((b + e) // 2).

    [b < e] whileTrue:[
	[b < end and:[(self at:b) < middleElement]] whileTrue:[b := b + 1].
	[e > begin and:[middleElement < (self at:e)]] whileTrue:[e := e - 1].

	(b <= e) ifTrue:[
	    (b == e) ifFalse:[
		temp1 := self at:b. temp2 := self at:e. 
		self at:b put:temp2. self at:e put:temp1
	    ].
	    b := b + 1.
	    e := e - 1
	]
    ].
    (begin < e) ifTrue:[self quickSortFrom:begin to:e].
    (b < end) ifTrue:[self quickSortFrom:b to:end]
!

quickSortFrom:inBegin to:inEnd sortBlock:sortBlock
    "actual quicksort worker for sort:-message"

    |begin   "{ Class: SmallInteger }"
     end     "{ Class: SmallInteger }"
     b       "{ Class: SmallInteger }"
     e       "{ Class: SmallInteger }"
     middleElement temp1 temp2 |

    begin := inBegin.   "/ this also does a type-check
    end := inEnd.

    b := begin.
    e := end.
    middleElement := self at:((b + e) // 2).

    [b < e] whileTrue:[
	[b < end and:[sortBlock value:(self at:b) value:middleElement]] whileTrue:[b := b + 1].
	[e > begin and:[sortBlock value:middleElement value:(self at:e)]] whileTrue:[e := e - 1].

	(b <= e) ifTrue:[
	    (b == e) ifFalse:[
		temp1 := self at:b. temp2 := self at:e. 
		self at:b put:temp2. self at:e put:temp1
	    ].
	    b := b + 1.
	    e := e - 1
	]
    ].
    (begin < e) ifTrue:[self quickSortFrom:begin to:e sortBlock:sortBlock].
    (b < end) ifTrue:[self quickSortFrom:b to:end sortBlock:sortBlock]
!

quickSortFrom:inBegin to:inEnd sortBlock:sortBlock policy:p
    "actual quicksort worker for sort:-message"

    |begin   "{ Class: SmallInteger }"
     end     "{ Class: SmallInteger }"
     b       "{ Class: SmallInteger }"
     e       "{ Class: SmallInteger }"
     middleElement temp1 temp2 |

    begin := inBegin.   "/ this also does a type-check
    end := inEnd.

    b := begin.
    e := end.
    middleElement := self at:((b + e) // 2).

    [b < e] whileTrue:[
        [b < end and:[sortBlock value:p value:(self at:b) value:middleElement]] whileTrue:[b := b + 1].
        [e > begin and:[sortBlock value:p value:middleElement value:(self at:e)]] whileTrue:[e := e - 1].

        (b <= e) ifTrue:[
            (b == e) ifFalse:[
                temp1 := self at:b. temp2 := self at:e. 
                self at:b put:temp2. self at:e put:temp1
            ].
            b := b + 1.
            e := e - 1
        ]
    ].
    (begin < e) ifTrue:[self quickSortFrom:begin to:e sortBlock:sortBlock policy:p].
    (b < end) ifTrue:[self quickSortFrom:b to:end sortBlock:sortBlock policy:p]

    "Created: / 27.10.1997 / 19:57:12 / cg"
!

quickSortFrom:inBegin to:inEnd sortBlock:sortBlock with:aCollection
    "actual quicksort worker for sort:with:-message"

    |begin   "{ Class: SmallInteger }"
     end     "{ Class: SmallInteger }"
     b       "{ Class: SmallInteger }"
     e       "{ Class: SmallInteger }"
     middleElement temp1 temp2 |

    begin := inBegin.   "/ this also does a type-check
    end := inEnd.

    b := begin.
    e := end.
    middleElement := self at:((b + e) // 2).

    [b < e] whileTrue:[
        [b < end and:[sortBlock value:(self at:b) value:middleElement]] whileTrue:[b := b + 1].
        [e > begin and:[sortBlock value:middleElement value:(self at:e)]] whileTrue:[e := e - 1].

        (b <= e) ifTrue:[
            (b == e) ifFalse:[
                temp1 := self at:b. temp2 := self at:e. 
                self at:b put:temp2. self at:e put:temp1.
                temp1 := aCollection at:b. temp2 := aCollection at:e. 
                aCollection at:b put:temp2. aCollection at:e put:temp1
            ].
            b := b + 1.
            e := e - 1
        ]
    ].
    (begin < e) ifTrue:[self quickSortFrom:begin to:e sortBlock:sortBlock with:aCollection].
    (b < end) ifTrue:[self quickSortFrom:b to:end sortBlock:sortBlock with:aCollection]

    "Modified: 5.1.1997 / 18:12:36 / cg"
!

quickSortFrom:inBegin to:inEnd with:aCollection
    "actual quicksort worker for sortWith-message"

    |begin   "{ Class: SmallInteger }"
     end     "{ Class: SmallInteger }"
     b       "{ Class: SmallInteger }"
     e       "{ Class: SmallInteger }"
     middleElement temp1 temp2 |

    begin := inBegin.   "/ this also does a type-check
    end := inEnd.

    b := begin.
    e := end.
    middleElement := self at:((b + e) // 2).

    [b < e] whileTrue:[
	[b < end and:[(self at:b) < middleElement]] whileTrue:[b := b + 1].
	[e > begin and:[middleElement < (self at:e)]] whileTrue:[e := e - 1].

	(b <= e) ifTrue:[
	    (b == e) ifFalse:[
		temp1 := self at:b. temp2 := self at:e. 
		self at:b put:temp2. self at:e put:temp1.
		temp1 := aCollection at:b. temp2 := aCollection at:e. 
		aCollection at:b put:temp2. aCollection at:e put:temp1
	    ].
	    b := b + 1.
	    e := e - 1
	]
    ].
    (begin < e) ifTrue:[self quickSortFrom:begin to:e with:aCollection].
    (b < end) ifTrue:[self quickSortFrom:b to:end with:aCollection]
!

randomizedQuickSortFrom:inBegin to:inEnd sortBlock:sortBlock with:aCollection
    "actual randomizedQuicksort worker for sort:with:-message.
     This exchanges a random element within the partition, to avoid
     the worst case O-square situation of quickSort.

     Notice, that this method has a much worse best- and average case
     runTime, due to the random number generation.
     The worst case of quickSort is encountered, if the choosen pivot
     element lies at either end of the partition, so that a split
     creates partitions of size 1 and (n-1).
     Since the middle element is choosen, this worst case is very unlikely
     to be encountered."

    |begin   "{ Class: SmallInteger }"
     end     "{ Class: SmallInteger }"
     b       "{ Class: SmallInteger }"
     e       "{ Class: SmallInteger }"
     rnd     "{ Class: SmallInteger }"
     middleElement temp1 temp2 |

    begin := inBegin.   "/ this also does a type-check
    end := inEnd.

    b := begin.
    e := end.

    "/ randomize that partition
    "/ by exchanging the first element with any random
    "/ element.
    rnd := Random nextIntegerBetween:b and:e.
    temp1 := self at:b. temp2 := self at:rnd.
    self at:b put:temp2. self at:rnd put:temp1.
    aCollection notNil ifTrue:[
        temp1 := aCollection at:b. temp2 := aCollection at:rnd. 
        aCollection at:b put:temp2. aCollection at:rnd put:temp1
    ].

    "/
    "/ now proceed as usual
    "/
    middleElement := self at:((b + e) // 2).

    [b < e] whileTrue:[
        [b < end and:[sortBlock value:(self at:b) value:middleElement]] whileTrue:[b := b + 1].
        [e > begin and:[sortBlock value:middleElement value:(self at:e)]] whileTrue:[e := e - 1].

        (b <= e) ifTrue:[
            (b == e) ifFalse:[
                temp1 := self at:b. temp2 := self at:e. 
                self at:b put:temp2. self at:e put:temp1.
                aCollection notNil ifTrue:[
                    temp1 := aCollection at:b. temp2 := aCollection at:e. 
                    aCollection at:b put:temp2. aCollection at:e put:temp1
                ].
            ].
            b := b + 1.
            e := e - 1
        ]
    ].
    (begin < e) ifTrue:[self randomizedQuickSortFrom:begin to:e sortBlock:sortBlock with:aCollection].
    (b < end) ifTrue:[self randomizedQuickSortFrom:b to:end sortBlock:sortBlock with:aCollection]

    "Created: 21.8.1997 / 18:13:35 / cg"
    "Modified: 21.8.1997 / 18:34:31 / cg"
!

randomizedQuickSortFrom:inBegin to:inEnd with:aCollection
    "actual randomizedQuicksort worker for randomizedSort:-message.
     This exchanges a random element within the partition, to avoid
     the worst case O-square situation of quickSort.

     Notice, that this method has a much worse best- and average case
     runTime, due to the random number generation.
     The worst case of quickSort is encountered, if the choosen pivot
     element lies at either end of the partition, so that a split
     creates partitions of size 1 and (n-1).
     Since the middle element is choosen, this worst case is very unlikely
     to be encountered."

    |begin   "{ Class: SmallInteger }"
     end     "{ Class: SmallInteger }"
     b       "{ Class: SmallInteger }"
     e       "{ Class: SmallInteger }"
     rnd     "{ Class: SmallInteger }"
     middleElement temp1 temp2 |

    begin := inBegin.   "/ this also does a type-check
    end := inEnd.

    b := begin.
    e := end.

    "/ randomize that partition
    "/ by exchanging the first element with any random
    "/ element.
    rnd := Random nextIntegerBetween:b and:e.
    temp1 := self at:b. temp2 := self at:rnd.
    self at:b put:temp2. self at:rnd put:temp1.
    aCollection notNil ifTrue:[
        temp1 := aCollection at:b. temp2 := aCollection at:rnd. 
        aCollection at:b put:temp2. aCollection at:rnd put:temp1
    ].

    "/
    "/ now proceed as usual
    "/
    middleElement := self at:((b + e) // 2).

    [b < e] whileTrue:[
        [b < end and:[(self at:b) < middleElement]] whileTrue:[b := b + 1].
        [e > begin and:[middleElement < (self at:e)]] whileTrue:[e := e - 1].

        (b <= e) ifTrue:[
            (b == e) ifFalse:[
                temp1 := self at:b. temp2 := self at:e. 
                self at:b put:temp2. self at:e put:temp1.
                aCollection notNil ifTrue:[
                    temp1 := aCollection at:b. temp2 := aCollection at:e. 
                    aCollection at:b put:temp2. aCollection at:e put:temp1
                ].
            ].
            b := b + 1.
            e := e - 1
        ]
    ].
    (begin < e) ifTrue:[self randomizedQuickSortFrom:begin to:e with:aCollection].
    (b < end) ifTrue:[self randomizedQuickSortFrom:b to:end with:aCollection]

    "Modified: 21.8.1997 / 18:30:19 / cg"
    "Created: 21.8.1997 / 18:34:23 / cg"
! !

!SequenceableCollection methodsFor:'queries'!

firstIndex
    "return the first elements index"

    ^ 1
!

isSequenceable
    "return true, if the receiver is some kind of sequenceableCollection"

    ^ true
!

isSequenceableCollection
    "OBSOLETE: use isSequenceable for ST-80 compatibility.
     This method is a historic leftover and will be removed."

    ^ true
!

keys
    "return a collection with all keys in the Smalltalk dictionary"

    |sz|

    sz := self size.
    sz == 0 ifTrue:[
	^ #()
    ].
    ^ 1 to:sz
!

lastIndex
    "return the last elements index"

    ^ self size
!

size
    "return the number of elements in the collection.
     concrete implementations must define this"

    ^ self subclassResponsibility
!

zeroIndex
    "return the index value which is returned when nothing
     is found in an indexOf* kind of search.
     ST-compatibility"

    ^ 0

    "Created: 14.2.1997 / 16:13:03 / cg"
! !

!SequenceableCollection methodsFor:'searching'!

detect:aBlock ifNone:exceptionBlock
    "find the first element, for which evaluation of the argument, aBlock
     return true; if none does so, return the evaluation of exceptionBlock

    reimplemented here for speed"

    |stop  "{ Class: SmallInteger }"
     element|

    stop := self size.
    1 to:stop do:[:index |
	element := self at:index.
	(aBlock value:element) ifTrue:[
	    ^ element
	].
    ].
    ^ exceptionBlock value

    "
     #(11 12 13 14) detect:[:n | n odd] ifNone:['sorry']    
     #(12 14 16 18) detect:[:n | n odd] ifNone:['sorry']     
    "
!

findFirst:aBlock
    "find the first element, for which evaluation of the argument, aBlock
     returns true; return its index or 0 if none detected."

    |stop  "{ Class: SmallInteger }" |

    stop := self size.
    1 to:stop do:[:index |
	(aBlock value:(self at:index)) ifTrue:[^ index].
    ].
    ^ 0

    "
     #(1 2 3 4 5 6) findFirst:[:x | (x >= 3)]
     #(1 2 3 4 5 6) findFirst:[:x | (x >= 3) and:[x even]]
     #(1 2 3 4 5 6) findFirst:[:x | (x >= 8)]           
     'one.two.three' findFirst:[:c | (c == $.)]
    "
!

findFirst:aBlock startingAt:startIndex
    "find the first element, for which evaluation of the argument, aBlock
     returns true; return its index or 0 if none detected."

    |start "{ Class: SmallInteger }"
     stop  "{ Class: SmallInteger }" |

    start := startIndex. "/ as a compiler hint
    stop := self size.
    start to:stop do:[:index |
	(aBlock value:(self at:index)) ifTrue:[^ index].
    ].
    ^ 0

    "
     #(1 4 3 4 3 6) findFirst:[:x | (x > 3)] startingAt:4
     'one.two.three' findFirst:[:c | (c == $.)] startingAt:5
    "
!

findLast:aBlock
    "find the last element, for which evaluation of the argument, aBlock
     returns true; return its index or 0 if none detected."

    |start "{ Class: SmallInteger }"|

    start := self size.
    start to:1 by:-1 do:[:index |
	(aBlock value:(self at:index)) ifTrue:[^ index].
    ].
    ^ 0

    "
     #(1 99 3 99 5 6) findLast:[:x | (x == 99)]
     'one.two.three' findLast:[:c | (c == $.)]
    "
!

findLast:aBlock startingAt:startIndex
    "find the last element, for which evaluation of the argument, aBlock
     returns true; return its index or 0 if none detected."

    |start "{ Class: SmallInteger }"|

    start := startIndex.
    start to:1 by:-1 do:[:index |
	(aBlock value:(self at:index)) ifTrue:[^ index].
    ].
    ^ 0

    "
     #(1 99 3 99 5 6) findLast:[:x | (x == 99)] startingAt:3
     'one.two.three' findLast:[:c | (c == $.)] startingAt:7
    "
!

identityIndexOf:anElement
    "search the collection for anElement using identity compare (i.e. ==);
     if found, return the index otherwise return 0."

    ^ self identityIndexOf:anElement startingAt:1

    "
     #(10 20 30 40 50 60 70) identityIndexOf:40
     #(10 20 30 40 50 60 70) identityIndexOf:40.0
     #(10 20 30 40 50 60 70) indexOf:40.0

    be careful:

     #(10 20 30 40.0 50 60 70) indexOf:40.0
     #(10 20 30 40.0 50 60 70) identityIndexOf:40.0

    "
!

identityIndexOf:anElement ifAbsent:exceptionBlock
    "search the collection for anElement using identity compare (i.e. ==);
     if found, return the index otherwise return the value of the
     exceptionBlock."

    |index|

    index := self identityIndexOf:anElement startingAt:1.
    (index == 0) ifTrue:[^ exceptionBlock value].
    ^ index

    "
     #(10 20 30 40 50 60 70) identityIndexOf:40  ifAbsent:['none']  
     #(10 20 30 40 50 60 70) identityIndexOf:35  ifAbsent:['none']  
     #(10 20 30 40 50 60 70) identityIndexOf:40.0 ifAbsent:['none'] 
     #(10 20 30 40 50 60 70) indexOf:40.0         ifAbsent:['none'] 
    "
!

identityIndexOf:anElement startingAt:start
    "search the collection for anElement, starting search at index start
     using identity compare  (i.e. ==);
     if found, return the index otherwise return 0."

    |startIndex "{ Class: SmallInteger }"
     stop       "{ Class: SmallInteger }" |

    startIndex := start.
    stop := self size.
    startIndex to:stop do:[:index |
	anElement == (self at:index) ifTrue:[^ index].
    ].
    ^ 0

    "
     #(10 20 30 40 10 20 30 40) identityIndexOf:40   startingAt:5
     #(10 20 30 40 10 20 30 40) identityIndexOf:40.0 startingAt:5
     #(10 20 30 40 10 20 30 40) indexOf:40.0         startingAt:5 
    "
!

identityIndexOf:anElement startingAt:start endingAt:stop
    "search the collection for anElement, starting the search at index start;
     ending at stop.
     If found (within the range), return the index, otherwise return 0.
     The comparison is done using = 
     (i.e. equality test - not identity test)."

    |startIndex "{ Class: SmallInteger }"
     stopIndex  "{ Class: SmallInteger }" |

    startIndex := start.
    stopIndex := self size min:stop.
    startIndex to:stopIndex do:[:index |
        anElement == (self at:index) ifTrue:[^ index].
    ].
    ^ 0

    "Created: 12.4.1996 / 18:23:07 / cg"
!

identityIndexOf:anElement startingAt:start ifAbsent:exceptionBlock
    "search the collection for anElement, starting search at index start;
     if found, return the index otherwise return the value of the
     exceptionBlock.
     This one searches for identical objects (i.e. ==)."

    |index|

    index := self identityIndexOf:anElement startingAt:start.
    (index == 0) ifTrue:[^ exceptionBlock value].
    ^ index

    "
     #(10 20 30 40 10) identityIndexOf:10 startingAt:3 ifAbsent:['none'] 
     #(10 20 30 40 10) identityIndexOf:35 startingAt:3 ifAbsent:['none'] 
    "
!

includes:anElement
    "return true if the collection contains anElement; false otherwise.
     Comparison is done using equality compare (i.e. =).
     See #includesIdentical: if identity is asked for."

    ((self indexOf:anElement startingAt:1) == 0) ifTrue:[^ false].
    ^ true

    "
     #(10 20 30 40 50 60 70) includes:99      
     #(10 20 30 40 50 60 70) includes:40     
     #(10 20 30 40 50 60 70) includes:40.0    
    "
!

includesIdentical:anElement
    "return true if the collection contains anElement; false otherwise.
     Comparison is done using identity compare (i.e. ==).
     See #includes: if equality is asked for."

    ((self identityIndexOf:anElement startingAt:1) == 0) ifTrue:[^ false].
    ^ true

    "
     #(10 20 30 40 50 60 70) includesIdentical:40     
     #(10 20 30 40 50 60 70) includesIdentical:40.0    
     #(10 20 30 40 50 60 70) includes:40     
     #(10 20 30 40 50 60 70) includes:40.0    

     be careful:

     #(10 20 30 40.0 50 60 70) includes:40.0    
     #(10 20 30 40.0 50 60 70) includesIdentical:40.0    
    "
!

indexOf:anElement
    "search the collection for anElement;
     if found, return the index otherwise return 0.
     The comparison is done using = 
     (i.e. equality test - not identity test)."

    ^ self indexOf:anElement startingAt:1

    "
     #(10 20 30 40 50 60 70) indexOf:40
     #(10 20 30 40 50 60 70) indexOf:40.0
    "
!

indexOf:anElement ifAbsent:exceptionBlock
    "search the collection for anElement;
     if found, return the index otherwise return the value of the
     exceptionBlock.
     The comparison is done using = 
     (i.e. equality test - not identity test)."

    |index|

    index := self indexOf:anElement startingAt:1.
    (index == 0) ifTrue:[^ exceptionBlock value].
    ^ index

    "
     #(10 20 30 40 10 20 30 40) indexOf:40   ifAbsent:['none'] 
     #(10 20 30 40 10 20 30 40) indexOf:40.0 ifAbsent:['none'] 
     #(10 20 30 40 10 20 30 40) indexOf:35   ifAbsent:['none'] 
    "
!

indexOf:elementToFind replaceWith:replacement startingAt:start stoppingAt:stop
    "search for the first occurence of elementToFind starting at start,
     stopping the search at stop. If found, replace the element by replacement
     and return the index.
     If not found, return 0."

    |idx|

    idx := self indexOf:elementToFind startingAt:start.
    ((idx > 0) and:[idx <= stop]) ifTrue:[
	self at:idx put:replacement.
	^ idx
    ].
    ^ 0

    "
     |a|

     a := #(10 20 30 40 50 60 70).
     (a indexOf:30 replaceWith:nil startingAt:1 stoppingAt:7) printNL.
     a printNL.
    "
!

indexOf:anElement startingAt:start
    "search the collection for anElement, starting the search at index start;
     if found, return the index otherwise return 0.
     The comparison is done using = 
     (i.e. equality test - not identity test)."

    |startIndex "{ Class: SmallInteger }"
     stop       "{ Class: SmallInteger }" |

    startIndex := start.
    stop := self size.
    startIndex to:stop do:[:index |
	anElement = (self at:index) ifTrue:[^ index].
    ].
    ^ 0

    "
     #(10 20 30 40 10 20 30 40) indexOf:40   startingAt:5  
     #(10 20 30 40 10 20 30 40) indexOf:40.0 startingAt:5  
    "
!

indexOf:anElement startingAt:start endingAt:stop
    "search the collection for anElement, starting the search at index start;
     ending at stop.
     If found (within the range), return the index, otherwise return 0.
     The comparison is done using = 
     (i.e. equality test - not identity test)."

    |startIndex "{ Class: SmallInteger }"
     stopIndex  "{ Class: SmallInteger }" |

    startIndex := start.
    stopIndex := self size min:stop.
    startIndex to:stopIndex do:[:index |
        anElement = (self at:index) ifTrue:[^ index].
    ].
    ^ 0

    "
     #(10 20 30 40 10 20 30 40) indexOf:40   startingAt:3 endingAt:99  
     #(10 20 30 40 10 20 30 40) indexOf:40   startingAt:3 endingAt:5  
     #(10 20 30 40 10 20 30 40) indexOf:40   startingAt:1 endingAt:3   
     #(10 20 30 40 10 20 30 40) indexOf:40.0 startingAt:5  
    "

    "Created: 12.4.1996 / 17:37:13 / cg"
!

indexOf:anElement startingAt:start ifAbsent:exceptionBlock
    "search the collection for anElement starting the search at index start;
     if found, return the index otherwise return the value of the
     exceptionBlock.
     The comparison is done using = 
     (i.e. equality test - not identity test)."

    |index|

    index := self indexOf:anElement startingAt:start.
    (index == 0) ifTrue:[^ exceptionBlock value].
    ^ index

    "
     #(10 20 30 40 10 20 30 40) indexOf:40   startingAt:5 ifAbsent:['none'] 
     #(10 20 30 40 10 20 30 40) indexOf:40.0 startingAt:5 ifAbsent:['none'] 
     #(10 20 30 40 10 20 30 40) indexOf:35   startingAt:5 ifAbsent:['none'] 
    "
!

indexOfAny:aCollection
    "search the collection for an element in aCollection.
     if found, return the index otherwise return 0.
     The comparison is done using = 
     (i.e. equality test - not identity test).

     Notice, that for big collections, the runtime of this search
     grows proportional to size(receiver) * size(aCollection).
     You may think about using other mechanisms (Sets, Dictionaries etc)."

    ^ self indexOfAny:aCollection startingAt:1

    "
     #(10 20 30 40 50 60 70) indexOfAny:#(40 30 50)
     #(10 20 30 40 50 60 70) indexOfAny:#(40.0 30.0 50)
    "
!

indexOfAny:aCollection startingAt:start
    "search the collection for an element in aCollection,
     starting the search at index start;
     if found, return the index otherwise return 0.
     The comparison is done using = 
     (i.e. equality test - not identity test).

     Notice, that for big collections, the runtime of this search
     grows proportional to size(receiver) * size(aCollection).
     You may think about using other mechanisms (Sets, Dictionaries etc)."


    |startIndex "{ Class: SmallInteger }"
     stop       "{ Class: SmallInteger }" |

    startIndex := start.
    stop := self size.
    startIndex to:stop do:[:index |
	(aCollection includes:(self at:index)) ifTrue:[^ index].
    ].
    ^ 0
    "
     #(10 20 30 40 10 20 30 40) indexOfAny:#(40 50 30) startingAt:5  
     #(10 20 30 40 10 20 30 40) indexOfAny:#(40.0 50 30.0) startingAt:5  
    "
!

indexOfAny:aCollection startingAt:start ifAbsent:exceptionBlock
    "search the collection for an element in aCollection,
     starting the search at index start;
     if found, return the index. If not, return the value returned by exceptionBlock.
     The comparison is done using = 
     (i.e. equality test - not identity test).

     Notice, that for big collections, the runtime of this search
     grows proportional to size(receiver) * size(aCollection).
     You may think about using other mechanisms (Sets, Dictionaries etc)."


    |val|

    val := self indexOfAny:aCollection startingAt:start .
    val == 0 ifTrue:[^ exceptionBlock value].
    ^ val.

    "
     #(10 20 30 40 10 20 30 40) indexOfAny:#(40 50 30) startingAt:5 ifAbsent:-1  
     #(10 20 30 40 10 20 30 40) indexOfAny:#(40.0 50 30.0) startingAt:5 ifAbsent:-1  
     #(10 20 30 40 10 20 30 40) indexOfAny:#(99 88 77) startingAt:5 ifAbsent:['oops']  
    "
!

indexOfSubCollection:aCollection
    "find a subcollection. 
     If found, return the index; if not found, return 0."

    ^ self indexOfSubCollection:aCollection startingAt:1 ifAbsent:0

    "
     #(1 2 3 4 5 6 7) indexOfSubCollection:#()  
     #(1 2 3 4 5 6 7) indexOfSubCollection:#(1)      
     #(1 2 3 4 5 6 7) indexOfSubCollection:#(1 2 3 4) 
     #(1 2 3 4 5 6 7) indexOfSubCollection:#(2 3 5)   
     #(1 2 3 2 3 4 5) indexOfSubCollection:#(2 3 4)  
     #(1 2 3 4 5 6 7) indexOfSubCollection:#(5 6 7)  
     #(1 2 3 4 5 6 7) indexOfSubCollection:#(5 6 8)  
     #(1 2 3 4 5 6 7) indexOfSubCollection:#(5 6 7 8)  
    "
!

indexOfSubCollection:aCollection ifAbsent:exceptionBlock
    "find a subcollection. If found, return the index;
     if not found, return the result of evaluating exceptionBlock."

    ^ self indexOfSubCollection:aCollection startingAt:1 ifAbsent:exceptionBlock
!

indexOfSubCollection:aCollection startingAt:startIndex
    "find a subcollection starting at index. 
     If found, return the index; if not found, return 0."

    ^ self indexOfSubCollection:aCollection startingAt:startIndex ifAbsent:0
!

indexOfSubCollection:aCollection startingAt:startIndex ifAbsent:exceptionBlock
    "find a subcollection, starting at index. If found, return the index;
     if not found, return the result of evaluating exceptionBlock.
     This is a q&d hack - not very efficient"

    |same first 
     mySize     "{Class: SmallInteger }"
     checkIndex "{Class: SmallInteger }"
     cmpIndex   "{Class: SmallInteger }"
     sz         "{Class: SmallInteger }"|

    mySize := self size.
    sz := aCollection size.
    sz == 0 ifTrue:[^ exceptionBlock value].
    first := aCollection at:1.

    checkIndex := startIndex - 1.    
    [true] whileTrue:[
	checkIndex := self indexOf:first startingAt:checkIndex+1.
	checkIndex == 0 ifTrue:[^ exceptionBlock value].

	(checkIndex + sz - 1) > mySize ifTrue:[
	    ^ 0
	].
	same := true.
	cmpIndex := 1.
	[same and:[cmpIndex <= sz]] whileTrue:[
	    (self at:checkIndex + cmpIndex - 1) = (aCollection at:cmpIndex) ifFalse:[
		same := false
	    ] ifTrue:[
		cmpIndex := cmpIndex + 1
	    ]
	].
	same ifTrue:[^ checkIndex].
    ]    

    "
     #(1 2 3 4 5 6 7) indexOfSubCollection:#()  startingAt:2 
     #(1 2 3 4 5 6 7) indexOfSubCollection:#(1) startingAt:2     
     #(1 2 1 2 1 2 3) indexOfSubCollection:#(1 2 3) startingAt:2 
     #(1 2 1 2 1 2 3) indexOfSubCollection:#(1 2)   startingAt:2  
     #(1 2 1 2 1 2 3) indexOfSubCollection:#(1 2)   startingAt:3  
     #(1 2 1 2 1 2 3) indexOfSubCollection:#(1 2)   startingAt:4 
    "
!

lastIndexOf:anElement
    "search the collection backwards for anElement;
     if found, return the index otherwise return 0.
     The comparison is done using = 
     (i.e. equality test - not identity test)."

    ^ self lastIndexOf:anElement startingAt:self size 

    "
     #(10 20 30 40 50 60 70) lastIndexOf:40    
     #(10 20 30 40 50 60 70) lastIndexOf:40.0  
     #(10 20 30 40 50 60 70) lastIndexOf:35    
     #(10 20 30 40 50 60 70) lastIndexOf:10    
    "
!

lastIndexOf:anElement ifAbsent:exceptionBlock
    "search the collection backwards for anElement;
     if found, return the index otherwise return the value of the
     exceptionBlock.
     The comparison is done using = 
     (i.e. equality test - not identity test)."

    |index|

    index := self lastIndexOf:anElement startingAt:self size.
    (index == 0) ifTrue:[^ exceptionBlock value].
    ^ index

    "
     #(10 20 30 40 10 20 30 40) lastIndexOf:40   ifAbsent:['none'] 
     #(10 20 30 40 10 20 30 40) lastIndexOf:40.0 ifAbsent:['none'] 
     #(10 20 30 40 10 20 30 40) lastIndexOf:35   ifAbsent:['none'] 
     #(10 20 30 40 10 20 30 40) lastIndexOf:10   ifAbsent:['none'] 
    "
!

lastIndexOf:anElement startingAt:start
    "search the collection backwards for anElement, starting the search at index start;
     if found, return the index otherwise return 0.
     The comparison is done using = 
     (i.e. equality test - not identity test)."

    |startIndex "{ Class: SmallInteger }"|

    startIndex := start.
    startIndex to:1 by:-1 do:[:index |
	anElement = (self at:index) ifTrue:[^ index].
    ].
    ^ 0

    "
     #(10 20 30 40 10 20 30 40) lastIndexOf:40   startingAt:8  
     #(10 20 30 40 10 20 30 40) lastIndexOf:40.0 startingAt:8  
     #(10 20 30 40 10 20 30 40) lastIndexOf:35   startingAt:8  
     #(10 20 30 40 10 20 30 40) lastIndexOf:10   startingAt:8  
     #(10 20 30 40 10 20 30 40) lastIndexOf:10   startingAt:4  
    "
!

lastIndexOf:anElement startingAt:start ifAbsent:exceptionBlock
    "search the collection backwards for anElement starting the search at 
     index start; if found, return the index 
     otherwise return the value of the exceptionBlock.
     The comparison is done using = 
     (i.e. equality test - not identity test)."

    |index|

    index := self lastIndexOf:anElement startingAt:start.
    (index == 0) ifTrue:[^ exceptionBlock value].
    ^ index

    "
     #(10 20 30 40 10 20 30 40) lastIndexOf:40   startingAt:8 ifAbsent:['none'] 
     #(10 20 30 40 10 20 30 40) lastIndexOf:40.0 startingAt:8 ifAbsent:['none'] 
     #(10 20 30 40 10 20 30 40) lastIndexOf:35   startingAt:8 ifAbsent:['none'] 
     #(10 20 30 40 10 20 30 40) lastIndexOf:10   startingAt:8 ifAbsent:['none'] 
    "
!

nextIndexOf:anElement from:start to:stop
    "search the collection for anElement, starting the search at index start,
     stopping at:stop;
     if found, return the index otherwise return 0.
     The comparison is done using = 
     (i.e. equality test - not identity test)."

    |startIndex "{ Class: SmallInteger }"
     stopIndex  "{ Class: SmallInteger }" |

    startIndex := start.
    stopIndex :=  stop.
    startIndex to:stop do:[:index |
	anElement = (self at:index) ifTrue:[^ index].
    ].
    ^ 0

    "
     #(10 20 30 40 10 20 30 40) nextIndexOf:40   from:2 to:6  
     #(10 20 30 40 10 20 30 40) nextIndexOf:40.0 from:2 to:6   
     #(10 20 30 40 10 20 30 40) nextIndexOf:35   from:2 to:6   
    "
!

nextIndexOf:anElement from:start to:stop ifAbsent:exceptionBlock
    "search the collection for anElement, starting the search at index start
     and stopping at stop;
     if found, return the index otherwise return the value of the
     exceptionBlock.
     The comparison is done using = 
     (i.e. equality test - not identity test)."

    |index|

    index := self nextIndexOf:anElement from:start to:stop.
    (index == 0) ifTrue:[^ exceptionBlock value].
    ^ index

    "
     #(10 20 30 40 10 20 30 40) nextIndexOf:40   from:2 to:6 ifAbsent:['none'] 
     #(10 20 30 40 10 20 30 40) nextIndexOf:40.0 from:2 to:6 ifAbsent:['none'] 
     #(10 20 30 40 10 20 30 40) nextIndexOf:35   from:2 to:6 ifAbsent:['none'] 
    "
!

prevIndexOf:anElement from:startSearchIndex to:endSearchIndex
    "search the collection for anElement, starting the search at index start;
     ending at stop, going in reverse direction.
     If found (within the range), return the index, otherwise return 0.
     The comparison is done using = 
     (i.e. equality test - not identity test)."

    |startIndex "{ Class: SmallInteger }"
     stopIndex  "{ Class: SmallInteger }" |

    startIndex := startSearchIndex.
    stopIndex := endSearchIndex max:1.
    startIndex to:stopIndex by:-1 do:[:index |
        anElement = (self at:index) ifTrue:[^ index].
    ].
    ^ 0

    "
     #(10 20 30 40 10 20 30 40) prevIndexOf:40   from:7 to:2  
     #(10 20 30 40 10 20 30 40) prevIndexOf:40.0 from:7 to:2  
     #(10 20 30 40 10 20 30 40) prevIndexOf:35   from:7 to:5  
    "

    "Modified: 25.1.1997 / 18:19:50 / cg"
!

prevIndexOf:anElement from:start to:stop ifAbsent:exceptionBlock
    "search the collection for anElement, starting the search at index start
     and stopping at stop, doing a reverse search;
     if found, return the index otherwise return the value of the
     exceptionBlock.
     The comparison is done using = 
     (i.e. equality test - not identity test)."

    |index|

    index := self prevIndexOf:anElement from:start to:stop.
    (index == 0) ifTrue:[^ exceptionBlock value].
    ^ index

    "
     #(10 20 30 40 10 20 30 40) prevIndexOf:40   from:7 to:2 ifAbsent:['none'] 
     #(10 20 30 40 10 20 30 40) prevIndexOf:40.0 from:7 to:2 ifAbsent:['none'] 
     #(10 20 30 40 10 20 30 40) prevIndexOf:35   from:7 to:5 ifAbsent:['none'] 
    "

    "Modified: 25.1.1997 / 18:18:18 / cg"
! !

!SequenceableCollection methodsFor:'sorting & reordering'!

randomizedSort
    "sort the collection inplace. The elements are compared using
     '<' i.e. they should offer a magnitude-like protocol.

     This uses the randomized quicksort algorithm, 
     which has a better worstCase behavior than quickSort
     (bit worse bestCase & averageCase behavior).
     See: Knuth or Cormen,Leiserson,Rivest pg. 163"

    |stop|

    stop := self size.
    (stop > 1) ifTrue:[
        self randomizedQuickSortFrom:1 to:stop sortBlock:[:a :b | a < b] with:nil
    ]

    "Modified: 21.8.1997 / 18:31:21 / cg"
!

randomizedSort:sortBlock
    "sort the collection inplace using the 2-arg block sortBlock
     for comparison. This allows any sort criteria to be implemented.

     This uses the randomized quicksort algorithm, 
     which has a better worstCase behavior than quickSort
     (bit worse bestCase & averageCase behavior).
     See: Knuth or Cormen,Leiserson,Rivest pg. 163"

    |stop|

    stop := self size.
    (stop > 1) ifTrue:[
        self randomizedQuickSortFrom:1 to:stop sortBlock:sortBlock with:nil
    ]

    "Created: 21.8.1997 / 18:16:50 / cg"
    "Modified: 21.8.1997 / 18:31:24 / cg"
!

randomizedSort:sortBlock with:aCollection
    "sort the collection inplace using the 2-arg block sortBlock
     for comparison. Also reorder the elements in aCollection.
     Use this, when you have a key collection to sort some other collection with.

     This uses the randomized quicksort algorithm, 
     which has a better worstCase behavior than quickSort
     (bit worse bestCase & averageCase behavior).
     See: Knuth or Cormen,Leiserson,Rivest pg. 163"

    |stop|

    stop := self size.
    (stop > 1) ifTrue:[
        self randomizedQuickSortFrom:1 to:stop sortBlock:sortBlock with:aCollection
    ]

    "Created: 21.8.1997 / 18:20:53 / cg"
    "Modified: 21.8.1997 / 18:31:27 / cg"
!

randomizedSortWith:aCollection
    "sort the receiver collection inplace, using '<' to compare elements.
     Also, the elements of aCollection are reordered with it.
     Use this, when you have a key collection to sort another collection with.

     This uses the randomized quicksort algorithm, 
     which has a better worstCase behavior than quickSort
     (bit worse bestCase & averageCase behavior).
     See: Knuth or Cormen,Leiserson,Rivest pg. 163"

    |stop|

    stop := self size.
    (stop > 1) ifTrue:[
        self randomizedQuickSortFrom:1 to:stop sortBlock:[:a :b | a < b] with:aCollection
    ]

    "Created: 21.8.1997 / 18:22:25 / cg"
    "Modified: 21.8.1997 / 18:31:31 / cg"
!

reverse
    "destructively reverse the order of the elements inplace"

    |lowIndex "{ Class: SmallInteger }"
     hiIndex  "{ Class: SmallInteger }"
     t1 t2|

    hiIndex := self size.
    lowIndex := 1.
    [lowIndex < hiIndex] whileTrue:[
	t1 := self at:lowIndex.  t2 := self at:hiIndex.
	self at:lowIndex put:t2.  self at:hiIndex put:t1.

	lowIndex := lowIndex + 1.
	hiIndex := hiIndex - 1
    ]

    "
     #(4 5 6 7 7) reverse
     #(1 4 7 10 2 5) asOrderedCollection reverse
    "
!

sort
    "sort the collection inplace. The elements are compared using
     '<' i.e. they should offer a magnitude-like protocol.

     The implementation uses the quicksort algorithm, which may not be
     the best possible for all situations (quickSort has O-square worst
     case behavior). See also #randomizedSort for a version with better
     worstCase behavior (but worse average & bestCase behavior)"

    |stop|

    stop := self size.
    (stop > 1) ifTrue:[
        self quickSortFrom:1 to:stop
    ]

    "
     #(1 16 7 98 3 19 4 0) sort

     |data|
     data := Random new next:100000.
     'random  ' print. (Time millisecondsToRun:[data sort]) printNL.
     'sorted  ' print. (Time millisecondsToRun:[data sort]) printNL.
     data reverse. 
     'reverse ' print. (Time millisecondsToRun:[data sort]) printNL.
    "

    "Modified: 21.8.1997 / 18:31:52 / cg"
!

sort:sortBlock
    "sort the collection inplace using the 2-arg block sortBlock
     for comparison. This allows any sort criteria to be implemented.

     The implementation uses the quicksort algorithm, which may not be
     the best possible for all situations (quickSort has O-square worst
     case behavior). See also #randomizedSort: for a version with better
     worstCase behavior (but worse average & bestCase behavior)"

    |stop|

    stop := self size.
    (stop > 1) ifTrue:[
        sortBlock numArgs == 3 ifTrue:[
            "/ TODO: pass a collating policy to aBlock
            self quickSortFrom:1 to:stop sortBlock:sortBlock policy:(StringCollationPolicy new)
        ] ifFalse:[
            self quickSortFrom:1 to:stop sortBlock:sortBlock
        ]
    ]

    "
     #(1 16 7 98 3 19 4 0) sort:[:a :b | a < b] 
     #(1 16 7 98 3 19 4 0) sort:[:a :b | a > b] 
     #(1 -1 16 -16 7 -7 98 -98) sort:[:a :b | a < b] 
     #(1 -1 16 -16 7 -7 98 -98) sort:[:a :b | a abs < b abs] 
    "

    "Modified: / 27.10.1997 / 20:03:22 / cg"
!

sort:sortBlock with:aCollection
    "sort the collection inplace using the 2-arg block sortBlock
     for comparison. Also reorder the elements in aCollection.
     Use this, when you have a key collection to sort some other collection with.

     The implementation uses the quicksort algorithm, which may not be
     the best possible for all situations (quickSort has O-square worst
     case behavior). See also #randomizedSort: for a version with better
     worstCase behavior (but worse average & bestCase behavior)"

    |stop|

    stop := self size.
    (stop > 1) ifTrue:[
        self quickSortFrom:1 to:stop sortBlock:sortBlock with:aCollection
    ]

    "
     |c1 c2|
     c1 := #(1 16 7 9).
     c2 := #('one' 'sixteen' 'seven' 'nine').
     c1 sort:[:a :b | a > b] with:c2.
     c1 printNL.
     c2 printNL
    "

    "Modified: 21.8.1997 / 18:32:11 / cg"
!

sortWith:aCollection
    "sort the receiver collection inplace, using '<' to compare elements.
     Also, the elements of aCollection are reordered with it.
     Use this, when you have a key collection to sort another collection with.

     The implementation uses the quicksort algorithm, which may not be
     the best possible for all situations (quickSort has O-square worst
     case behavior). See also #randomizedSort: for a version with better
     worstCase behavior (but worse average & bestCase behavior)"

    |stop|

    stop := self size.
    (stop > 1) ifTrue:[
        self quickSortFrom:1 to:stop with:aCollection
    ]

    "Modified: 21.8.1997 / 18:32:21 / cg"
!

topologicalSort:sortBlock
    "sort the collection inplace using a sloooow sort algorithm.
     This algorithm has O-square runtime behavior and should be used only
     in special situations.
     It compares all elements, thus can be used when a>b, b>c does NOT imply
     a>c (for example, to sort classes by inheritance)

     In all other situations, use #sort; which implements the quicksort algorithm.
    "

    |smallestIndex "{ Class: SmallInteger }"
     end           "{ Class: SmallInteger }"
     index2        "{ Class: SmallInteger }"
     smallest thisOne|

    "this is just a q&d hack - there must be better implementations for this ;-)"

    end := self size.
    1 to:end-1 do:[:index |
        smallest := self at:index.
        smallestIndex := index.

        index2 := index + 1.
        [
            thisOne := self at:index2.
            (index2 ~~ smallestIndex 
            and:[sortBlock value:thisOne value:smallest]) ifTrue:[
                smallestIndex := index2.
                smallest := thisOne.
                index2 := index + 1.
            ] ifFalse:[
                index2 := index2 + 1..
            ].
        ] doWhile:[index2 <= end].

"/        (index + 1) to:end do:[:index2 |
"/            thisOne := self at:index2.
"/            (sortBlock value:thisOne value:smallest) ifTrue:[
"/                smallestIndex := index2.
"/                smallest := thisOne
"/            ].
"/        ].

        (smallestIndex ~~ index) ifTrue:[
            thisOne := self at:index.
            self at:index put:smallest.
            self at:smallestIndex put:thisOne
        ].
    ]

    "
     #(1 16 7 98 3 19 4 0) topologicalSort:[:a :b | a < b]   
     #(1 16 7 98 3 19 4 0) sort:[:a :b | a < b]              
     Smalltalk allClasses asArray topologicalSort:[:a :b | b isSubclassOf:a] 
     Smalltalk allClasses asArray sort:[:a :b | b isSubclassOf:a] 
    "

    "Modified: 21.8.1997 / 18:30:49 / cg"
! !

!SequenceableCollection class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libbasic/Attic/SeqColl.st,v 1.89 1997-10-28 18:57:00 cg Exp $'
! !
